﻿using System;
using System.Data.Common;
using System.IO;
using System.Threading;
using System.Threading.Tasks;

namespace Wunnell.Extensions
{
    /// <summary>
    /// Contains methods that extend the <see cref="DbDataReader"/> class.
    /// </summary>
    public static class DataReaderExtensions
    {
        /// <summary>
        /// Synchronously gets the value of the specified column as a type.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the value to be returned.
        /// </typeparam>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="ordinal">
        /// The zero-based column ordinal.
        /// </param>
        /// <returns>
        /// The value of the specified column. 
        /// </returns>
        /// <remarks>
        /// This method also supports nullable value types, which <see cref="DbDataReader.GetFieldValue{T}">GetFieldValue</see> does not.
        /// </remarks>
        public static T Field<T>(this DbDataReader source, int ordinal)
        {
            return source.IsDBNull(ordinal)
                       ? default(T)
                       : source.GetFieldValue<T>(ordinal);
        }

        /// <summary>
        /// Synchronously gets the value of the specified column as a type.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the value to be returned.
        /// </typeparam>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column. 
        /// </returns>
        /// <remarks>
        /// This method also supports nullable value types, which <see cref="GetFieldValue{T}">GetFieldValue</see> does not.
        /// </remarks>
        public static T Field<T>(this DbDataReader source, string name)
        {
            return source.Field<T>(source.GetOrdinal(name));
        }

        /// <summary>
        /// Asynchronously gets the value of the specified column as a type.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the value to be returned.
        /// </typeparam>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="ordinal">
        /// The zero-based column ordinal.
        /// </param>
        /// <returns>
        /// The value of the specified column. 
        /// </returns>
        /// <remarks>
        /// This method also supports nullable value types, which <see cref="DbDataReader.GetFieldValueAsync{T}(int)">GetFieldValueAsync</see> does not.
        /// </remarks>
        public static Task<T> FieldAsync<T>(this DbDataReader source, int ordinal)
        {
            return source.FieldAsync<T>(ordinal, CancellationToken.None);
        }

        /// <summary>
        /// Asynchronously gets the value of the specified column as a type.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the value to be returned.
        /// </typeparam>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column. 
        /// </returns>
        /// <remarks>
        /// This method also supports nullable value types, which <see cref="GetFieldValueAsync{T}(DbDataReader,string)">GetFieldValueAsync</see> does not.
        /// </remarks>
        public static Task<T> FieldAsync<T>(this DbDataReader source, string name)
        {
            return source.FieldAsync<T>(name, CancellationToken.None);
        }

        /// <summary>
        /// Asynchronously gets the value of the specified column as a type.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the value to be returned.
        /// </typeparam>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="ordinal">
        /// The zero-based column ordinal.
        /// </param>
        /// <param name="cancellationToken">
        /// The cancellation instruction, which propagates a notification that operations should be canceled. This does not guarantee the cancellation. A setting of <b>CancellationToken.None</b> makes this method equivalent to <see cref="GetFieldValueAsync{T}(DbDataReader,string)">GetFieldValueAsync</see>. The returned task must be marked as cancelled.
        /// </param>
        /// <returns>
        /// The value of the specified column. 
        /// </returns>
        /// <remarks>
        /// This method also supports nullable value types, which <see cref="DbDataReader.GetFieldValueAsync{T}(int,CancellationToken)">GetFieldValueAsync</see> does not.
        /// </remarks>
        public static Task<T> FieldAsync<T>(this DbDataReader source, int ordinal, CancellationToken cancellationToken)
        {
            if (cancellationToken.IsCancellationRequested)
            {
                return Task.FromCanceled<T>(cancellationToken);
            }

            try
            {
                return source.IsDBNull(ordinal)
                           ? Task.FromResult(default(T))
                           : Task.FromResult(source.GetFieldValue<T>(ordinal));
            }
            catch (Exception e)
            {
                return Task.FromException<T>(e);
            }
        }

        /// <summary>
        /// Asynchronously gets the value of the specified column as a type.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the value to be returned.
        /// </typeparam>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <param name="cancellationToken">
        /// The cancellation instruction, which propagates a notification that operations should be canceled. This does not guarantee the cancellation. A setting of <b>CancellationToken.None</b> makes this method equivalent to <see cref="GetFieldValueAsync{T}(DbDataReader,string)">GetFieldValueAsync</see>. The returned task must be marked as cancelled.
        /// </param>
        /// <returns>
        /// The value of the specified column. 
        /// </returns>
        /// <remarks>
        /// This method also supports nullable value types, which <see cref="GetFieldValueAsync{T}(DbDataReader,string,CancellationToken)">GetFieldValueAsync</see> does not.
        /// </remarks>
        public static Task<T> FieldAsync<T>(this DbDataReader source, string name, CancellationToken cancellationToken)
        {
            return source.FieldAsync<T>(source.GetOrdinal(name), cancellationToken);
        }

        /// <summary>
        /// Gets the value of the specified column as a Boolean.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static bool GetBoolean(this DbDataReader source, string name)
        {
            return source.GetBoolean(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as a byte.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static byte GetByte(this DbDataReader source, string name)
        {
            return source.GetByte(source.GetOrdinal(name));
        }

        /// <summary>
        /// Reads a stream of bytes from the specified column, starting at location indicated by <i>dataOffset</i>, into the buffer, starting at the location indicated by <i>bufferOffset</i>.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <param name="dataOffset">
        /// The index within the row from which to begin the read operation.
        /// </param>
        /// <param name="buffer">
        /// The buffer into which to copy the data.
        /// </param>
        /// <param name="bufferOffset">
        /// The index with the buffer to which the data will be copied.
        /// </param>
        /// <param name="length">
        /// The maximum number of bytes to read.
        /// </param>
        /// <returns>
        /// The actual number of bytes read.
        /// </returns>
        public static long GetBytes(this DbDataReader source,
                                    string name,
                                    long dataOffset,
                                    byte[] buffer,
                                    int bufferOffset,
                                    int length)
        {
            return source.GetBytes(source.GetOrdinal(name),
                                   dataOffset,
                                   buffer,
                                   bufferOffset,
                                   length);
        }

        /// <summary>
        /// Gets the value of the specified column as a single character.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static char GetChar(this DbDataReader source, string name)
        {
            return source.GetChar(source.GetOrdinal(name));
        }

        /// <summary>
        /// Reads a stream of characters from the specified column, starting at location indicated by <i>dataOffset</i>, into the buffer, starting at the location indicated by <i>bufferOffset</i>.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <param name="dataOffset">
        /// The index within the row from which to begin the read operation.
        /// </param>
        /// <param name="buffer">
        /// The buffer into which to copy the data.
        /// </param>
        /// <param name="bufferOffset">
        /// The index with the buffer to which the data will be copied.
        /// </param>
        /// <param name="length">
        /// The maximum number of characters to read.
        /// </param>
        /// <returns>
        /// The actual number of characters read.
        /// </returns>
        public static long GetChars(this DbDataReader source,
                                    string name,
                                    long dataOffset,
                                    char[] buffer,
                                    int bufferOffset,
                                    int length)
        {
            return source.GetChars(source.GetOrdinal(name),
                                   dataOffset,
                                   buffer,
                                   bufferOffset,
                                   length);
        }

        /// <summary>
        /// Returns a <see cref="DbDataReader"/> object for the requested column name.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// A <see cref="DbDataReader"/> object.
        /// </returns>
        public static DbDataReader GetData(this DbDataReader source, string name)
        {
            return source.GetData(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets name of the data type of the specified column.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// A string representing the name of the data type.
        /// </returns>
        public static string GetDataTypeName(this DbDataReader source, string name)
        {
            return source.GetDataTypeName(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as a <see cref="Decimal"/> object.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static DateTime GetDateTime(this DbDataReader source, string name)
        {
            return source.GetDateTime(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as a <see cref="DateTime"/> object.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static decimal GetDecimal(this DbDataReader source, string name)
        {
            return source.GetDecimal(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as a double-precision floating point number.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static double GetDouble(this DbDataReader source, string name)
        {
            return source.GetDouble(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the data type of the specified column.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The data type of the specified column.
        /// </returns>
        public static Type GetFieldType(this DbDataReader source, string name)
        {
            return source.GetFieldType(source.GetOrdinal(name));
        }

        /// <summary>
        /// Synchronously gets the value of the specified column as a type.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the value to be returned.
        /// </typeparam>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column. 
        /// </returns>
        public static T GetFieldValue<T>(this DbDataReader source, string name)
        {
            return source.GetFieldValue<T>(source.GetOrdinal(name));
        }

        /// <summary>
        /// Asynchronously gets the value of the specified column as a type.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the value to be returned.
        /// </typeparam>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column. 
        /// </returns>
        public static Task<T> GetFieldValueAsync<T>(this DbDataReader source, string name)
        {
            return source.GetFieldValueAsync<T>(source.GetOrdinal(name));
        }

        /// <summary>
        /// Asynchronously gets the value of the specified column as a type.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the value to be returned.
        /// </typeparam>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <param name="cancellationToken">
        /// The cancellation instruction, which propagates a notification that operations should be canceled. This does not guarantee the cancellation. A setting of <b>CancellationToken.None</b> makes this method equivalent to <see cref="GetFieldValueAsync{T}(DbDataReader,string)">GetFieldValueAsync</see>. The returned task must be marked as cancelled.
        /// </param>
        /// <returns>
        /// The value of the specified column. 
        /// </returns>
        public static Task<T> GetFieldValueAsync<T>(this DbDataReader source, string name, CancellationToken cancellationToken)
        {
            return source.GetFieldValueAsync<T>(source.GetOrdinal(name), cancellationToken);
        }

        /// <summary>
        /// Gets the value of the specified column as a single-precision floating point number.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static float GetFloat(this DbDataReader source, string name)
        {
            return source.GetFloat(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as a globally-unique identifier (GUID).
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static Guid GetGuid(this DbDataReader source, string name)
        {
            return source.GetGuid(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as a 16-bit signed integer.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static short GetInt16(this DbDataReader source, string name)
        {
            return source.GetInt16(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as a 32-bit signed integer.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static int GetInt32(this DbDataReader source, string name)
        {
            return source.GetInt32(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as a 64-bit signed integer.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static long GetInt64(this DbDataReader source, string name)
        {
            return source.GetInt64(source.GetOrdinal(name));
        }

        /// <summary>
        /// Returns the provider-specific field type of the specified column.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The <see cref="Type"/> object that describes the data type of the specified column.
        /// </returns>
        public static Type GetProviderSpecificFieldType(this DbDataReader source, string name)
        {
            return source.GetProviderSpecificFieldType(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as an instance of <see cref="Object"/>.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static object GetProviderSpecificValue(this DbDataReader source, string name)
        {
            return source.GetProviderSpecificValue(source.GetOrdinal(name));
        }

        /// <summary>
        /// Retrieves data as a <see cref="Stream"/>.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The returned object.
        /// </returns>
        public static Stream GetStream(this DbDataReader source, string name)
        {
            return source.GetStream(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as an instance of <see cref="String"/>.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static string GetString(this DbDataReader source, string name)
        {
            return source.GetString(source.GetOrdinal(name));
        }

        /// <summary>
        /// Retrieves data as a <see cref="TextReader"/>.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The returned object.
        /// </returns>
        public static TextReader GetTextReader(this DbDataReader source, string name)
        {
            return source.GetTextReader(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets the value of the specified column as an instance of <see cref="Object"/>.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// The value of the specified column.
        /// </returns>
        public static object GetValue(this DbDataReader source, string name)
        {
            return source.GetValue(source.GetOrdinal(name));
        }

        /// <summary>
        /// Gets a value that indicates whether the column contains nonexistent or missing values.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// <b>true</b> if the specified column is equivalent to <see cref="DBNull"/>; otherwise <b>false</b>.
        /// </returns>
        public static bool IsDBNull(this DbDataReader source, string name)
        {
            return source.IsDBNull(source.GetOrdinal(name));
        }

        /// <summary>
        /// An asynchronous version of <see cref="IsDBNull"/>, which gets a value that indicates whether the column contains non-existent or missing values.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <returns>
        /// <b>true</b> if the specified column is equivalent to <see cref="DBNull"/>; otherwise <b>false</b>.
        /// </returns>
        public static Task<bool> IsDBNullAsync(this DbDataReader source, string name)
        {
            return source.IsDBNullAsync(source.GetOrdinal(name));
        }

        /// <summary>
        /// An asynchronous version of <see cref="IsDBNull"/>, which gets a value that indicates whether the column contains non-existent or missing values. Optionally, sends a notification that operations should be cancelled.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="name">
        /// The name of the column.
        /// </param>
        /// <param name="cancellationToken">
        /// The cancellation instruction, which propagates a notification that operations should be canceled. This does not guarantee the cancellation. A setting of <b>CancellationToken.None</b> makes this method equivalent to <see cref="IsDBNullAsync(DbDataReader,string)">IsDBNullAsync</see>. The returned task must be marked as cancelled.
        /// </param>
        /// <returns>
        /// <b>true</b> if the specified column is equivalent to <see cref="DBNull"/>; otherwise <b>false</b>.
        /// </returns>
        public static Task<bool> IsDBNullAsync(this DbDataReader source, string name, CancellationToken cancellationToken)
        {
            return source.IsDBNullAsync(source.GetOrdinal(name), cancellationToken);
        }
    }
}
