﻿Imports System.Data.Common
Imports System.IO
Imports System.Runtime.CompilerServices
Imports System.Threading

''' <summary>
''' Contains methods that extend the <see cref="DbDataReader"/> class.
''' </summary>
Public Module DataReaderExtensions

    ''' <summary>
    ''' Synchronously gets the value of the specified column as a type.
    ''' </summary>
    ''' <typeparam name="T">
    ''' The type of the value to be returned.
    ''' </typeparam>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="ordinal">
    ''' The zero-based column ordinal.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column. 
    ''' </returns>
    ''' <remarks>
    ''' This method also supports nullable value types, which <see cref="DbDataReader.GetFieldValue(Of T)">GetFieldValue</see> does not.
    ''' </remarks>
    <Extension>
    Public Function Field(Of T)(source As DbDataReader, ordinal As Integer) As T
        Return If(source.IsDBNull(ordinal),
                  Nothing,
                  source.GetFieldValue(Of T)(ordinal))
    End Function

    ''' <summary>
    ''' Synchronously gets the value of the specified column as a type.
    ''' </summary>
    ''' <typeparam name="T">
    ''' The type of the value to be returned.
    ''' </typeparam>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column. 
    ''' </returns>
    ''' <remarks>
    ''' This method also supports nullable value types, which <see cref="GetFieldValue(Of T)">GetFieldValue</see> does not.
    ''' </remarks>
    <Extension>
    Public Function Field(Of T)(source As DbDataReader, name As String) As T
        Return source.Field(Of T)(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Asynchronously gets the value of the specified column as a type.
    ''' </summary>
    ''' <typeparam name="T">
    ''' The type of the value to be returned.
    ''' </typeparam>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="ordinal">
    ''' The zero-based column ordinal.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column. 
    ''' </returns>
    ''' <remarks>
    ''' This method also supports nullable value types, which <see cref="DbDataReader.GetFieldValueAsync(Of T)(Integer)">GetFieldValueAsync</see> does Not.
    ''' </remarks>
    <Extension>
    Public Function FieldAsync(Of T)(source As DbDataReader, ordinal As Integer) As Task(Of T)
        Return source.FieldAsync(Of T)(ordinal, CancellationToken.None)
    End Function

    ''' <summary>
    ''' Asynchronously gets the value of the specified column as a type.
    ''' </summary>
    ''' <typeparam name="T">
    ''' The type of the value to be returned.
    ''' </typeparam>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column. 
    ''' </returns>
    ''' <remarks>
    ''' This method also supports nullable value types, which <see cref="GetFieldValueAsync(Of T)(DbDataReader,String)">GetFieldValueAsync</see> does Not.
    ''' </remarks>
    <Extension>
    Public Function FieldAsync(Of T)(source As DbDataReader, name As String) As Task(Of T)
        Return source.FieldAsync(Of T)(name, CancellationToken.None)
    End Function

    ''' <summary>
    ''' Asynchronously gets the value of the specified column as a type.
    ''' </summary>
    ''' <typeparam name="T">
    ''' The type of the value to be returned.
    ''' </typeparam>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="ordinal">
    ''' The zero-based column ordinal.
    ''' </param>
    ''' <param name="cancellationToken">
    ''' The cancellation instruction, which propagates a notification that operations should be canceled. This does Not guarantee the cancellation. A setting of <b>CancellationToken.None</b> makes this method equivalent to <see cref="GetFieldValueAsync(Of T)(DbDataReader,String)">GetFieldValueAsync</see>. The returned task must be marked as cancelled.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column. 
    ''' </returns>
    ''' <remarks>
    ''' This method also supports nullable value types, which <see cref="DbDataReader.GetFieldValueAsync(Of T)(Integer,CancellationToken)">GetFieldValueAsync</see> does Not.
    ''' </remarks>
    <Extension>
    Public Function FieldAsync(Of T)(source As DbDataReader, ordinal As Integer, cancellationToken As CancellationToken) As Task(Of T)
        If cancellationToken.IsCancellationRequested Then
            Return Task.FromCanceled(Of T)(cancellationToken)
        End If

        Try
            Return If(source.IsDBNull(ordinal),
                      Task.FromResult(Of T)(Nothing),
                      Task.FromResult(source.GetFieldValue(Of T)(ordinal)))
        Catch ex As Exception
            Return Task.FromException(Of T)(ex)
        End Try
    End Function

    ''' <summary>
    ''' Asynchronously gets the value of the specified column as a type.
    ''' </summary>
    ''' <typeparam name="T">
    ''' The type of the value to be returned.
    ''' </typeparam>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <param name="cancellationToken">
    ''' The cancellation instruction, which propagates a notification that operations should be canceled. This does Not guarantee the cancellation. A setting of <b>CancellationToken.None</b> makes this method equivalent to <see cref="GetFieldValueAsync(Of T)(DbDataReader,String)">GetFieldValueAsync</see>. The returned task must be marked as cancelled.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column. 
    ''' </returns>
    ''' <remarks>
    ''' This method also supports nullable value types, which <see cref="GetFieldValueAsync(Of T)(DbDataReader,String,CancellationToken)">GetFieldValueAsync</see> does Not.
    ''' </remarks>
    <Extension>
    Public Function FieldAsync(Of T)(source As DbDataReader, name As String, cancellationToken As CancellationToken) As Task(Of T)
        Return source.FieldAsync(Of T)(source.GetOrdinal(name), cancellationToken)
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a Boolean.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetBoolean(source As DbDataReader, name As String) As Boolean
        Return source.GetBoolean(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a byte.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetByte(source As DbDataReader, name As String) As Byte
        Return source.GetByte(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Reads a stream of bytes from the specified column, starting at location indicated by <i>dataOffset</i>, into the buffer, starting at the location indicated by <i>bufferOffset</i>.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <param name="dataOffset">
    ''' The index within the row from which to begin the read operation.
    ''' </param>
    ''' <param name="buffer">
    ''' The buffer into which to copy the data.
    ''' </param>
    ''' <param name="bufferOffset">
    ''' The index with the buffer to which the data will be copied.
    ''' </param>
    ''' <param name="length">
    ''' The maximum number of bytes to read.
    ''' </param>
    ''' <returns>
    ''' The actual number of bytes read.
    ''' </returns>
    <Extension>
    Public Function GetBytes(source As DbDataReader,
                             name As String,
                             dataOffset As Long,
                             buffer As Byte(),
                             bufferOffset As Integer,
                             length As Integer) As Long
        Return source.GetBytes(source.GetOrdinal(name),
                               dataOffset,
                               buffer,
                               bufferOffset,
                               length)
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a single character.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetChar(source As DbDataReader, name As String) As Char
        Return source.GetChar(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Reads a stream of characters from the specified column, starting at location indicated by <i>dataOffset</i>, into the buffer, starting at the location indicated by <i>bufferOffset</i>.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <param name="dataOffset">
    ''' The index within the row from which to begin the read operation.
    ''' </param>
    ''' <param name="buffer">
    ''' The buffer into which to copy the data.
    ''' </param>
    ''' <param name="bufferOffset">
    ''' The index with the buffer to which the data will be copied.
    ''' </param>
    ''' <param name="length">
    ''' The maximum number of characters to read.
    ''' </param>
    ''' <returns>
    ''' The actual number of characters read.
    ''' </returns>
    <Extension>
    Public Function GetChars(source As DbDataReader,
                             name As String,
                             dataOffset As Long,
                             buffer As Char(),
                             bufferOffset As Integer,
                             length As Integer) As Long
        Return source.GetChars(source.GetOrdinal(name),
                               dataOffset,
                               buffer,
                               bufferOffset,
                               length)
    End Function

    ''' <summary>
    ''' Returns a <see cref="DbDataReader"/> object for the requested column name.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' A <see cref="DbDataReader"/> object.
    ''' </returns>
    <Extension>
    Public Function GetData(source As DbDataReader, name As String) As DbDataReader
        Return source.GetData(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets name of the data type of the specified column.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' A string representing the name of the data type.
    ''' </returns>
    <Extension>
    Public Function GetDataTypeName(source As DbDataReader, name As String) As String
        Return source.GetDataTypeName(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a <see cref="Decimal"/> object.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetDateTime(source As DbDataReader, name As String) As Date
        Return source.GetDateTime(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a <see cref="DateTime"/> object.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetDecimal(source As DbDataReader, name As String) As Decimal
        Return source.GetDecimal(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a double-precision floating point number.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetDouble(source As DbDataReader, name As String) As Double
        Return source.GetDouble(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the data type of the specified column.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The data type of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetFieldType(source As DbDataReader, name As String) As Type
        Return source.GetFieldType(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Synchronously gets the value of the specified column as a type.
    ''' </summary>
    ''' <typeparam name="T">
    ''' The type of the value to be returned.
    ''' </typeparam>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column. 
    ''' </returns>
    <Extension>
    Public Function GetFieldValue(Of T)(source As DbDataReader, name As String) As T
        Return source.GetFieldValue(Of T)(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Asynchronously gets the value of the specified column as a type.
    ''' </summary>
    ''' <typeparam name="T">
    ''' The type of the value to be returned.
    ''' </typeparam>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column. 
    ''' </returns>
    <Extension>
    Public Function GetFieldValueAsync(Of T)(source As DbDataReader, name As String) As Task(Of T)
        Return source.GetFieldValueAsync(Of T)(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Asynchronously gets the value of the specified column as a type.
    ''' </summary>
    ''' <typeparam name="T">
    ''' The type of the value to be returned.
    ''' </typeparam>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <param name="cancellationToken">
    ''' The cancellation instruction, which propagates a notification that operations should be canceled. This does Not guarantee the cancellation. A setting of <b>CancellationToken.None</b> makes this method equivalent to <see cref="GetFieldValueAsync(Of T)(DbDataReader,String)">GetFieldValueAsync</see>. The returned task must be marked as cancelled.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column. 
    ''' </returns>
    <Extension>
    Public Function GetFieldValueAsync(Of T)(source As DbDataReader, name As String, cancellationToken As CancellationToken) As Task(Of T)
        Return source.GetFieldValueAsync(Of T)(source.GetOrdinal(name), cancellationToken)
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a single-precision floating point number.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetFloat(source As DbDataReader, name As String) As Single
        Return source.GetFloat(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a globally-unique identifier (GUID).
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetGuid(source As DbDataReader, name As String) As Guid
        Return source.GetGuid(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a 16-bit signed integer.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetInt16(source As DbDataReader, name As String) As Short
        Return source.GetInt16(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a 32-bit signed integer.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetInt32(source As DbDataReader, name As String) As Integer
        Return source.GetInt32(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as a 64-bit signed integer.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetInt64(source As DbDataReader, name As String) As Long
        Return source.GetInt64(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Returns the provider-specific field type of the specified column.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The <see cref="Type"/> object that describes the data type of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetProviderSpecificFieldType(source As DbDataReader, name As String) As Type
        Return source.GetProviderSpecificFieldType(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as an instance of <see cref="Object"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetProviderSpecificValue(source As DbDataReader, name As String) As Object
        Return source.GetProviderSpecificValue(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Retrieves data as a <see cref="Stream"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The returned object.
    ''' </returns>
    <Extension>
    Public Function GetStream(source As DbDataReader, name As String) As Stream
        Return source.GetStream(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as an instance of <see cref="String"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetString(source As DbDataReader, name As String) As String
        Return source.GetString(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Retrieves data as a <see cref="TextReader"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The returned object.
    ''' </returns>
    <Extension>
    Public Function GetTextReader(source As DbDataReader, name As String) As TextReader
        Return source.GetTextReader(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets the value of the specified column as an instance of <see cref="Object"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' The value of the specified column.
    ''' </returns>
    <Extension>
    Public Function GetValue(source As DbDataReader, name As String) As Object
        Return source.GetValue(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' Gets a value that indicates whether the column contains nonexistent Or missing values.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the specified column Is equivalent to <see cref="DBNull"/>; otherwise <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsDBNull(source As DbDataReader, name As String) As Boolean
        Return source.IsDBNull(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' An asynchronous version of <see cref="IsDBNull"/>, which gets a value that indicates whether the column contains non-existent Or missing values.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the specified column Is equivalent to <see cref="DBNull"/>; otherwise <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsDBNullAsync(source As DbDataReader, name As String) As Task(Of Boolean)
        Return source.IsDBNullAsync(source.GetOrdinal(name))
    End Function

    ''' <summary>
    ''' An asynchronous version of <see cref="IsDBNull"/>, which gets a value that indicates whether the column contains non-existent Or missing values. Optionally, sends a notification that operations should be cancelled.
    ''' </summary>
    ''' <param name="source">
    ''' The input <see cref="DbDataReader"/>, which acts as the <b>this</b> instance for the extension method.
    ''' </param>
    ''' <param name="name">
    ''' The name of the column.
    ''' </param>
    ''' <param name="cancellationToken">
    ''' The cancellation instruction, which propagates a notification that operations should be canceled. This does Not guarantee the cancellation. A setting of <b>CancellationToken.None</b> makes this method equivalent to <see cref="IsDBNullAsync(DbDataReader,String)">IsDBNullAsync</see>. The returned task must be marked as cancelled.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the specified column Is equivalent to <see cref="DBNull"/>; otherwise <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsDBNullAsync(source As DbDataReader, name As String, cancellationToken As CancellationToken) As Task(Of Boolean)
        Return source.IsDBNullAsync(source.GetOrdinal(name), cancellationToken)
    End Function

End Module
