﻿using System;

namespace Wunnell.Extensions
{
    /// <summary>
    /// Contains methods that extend the <see cref="DateTime"/> structure.
    /// </summary>
    public static class DateTimeExtensions
    {
        /// <summary>
        /// Gets a value indicating whether a <see cref="DateTime"/> value represents a week day.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DateTime"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <returns>
        /// <b>true</b> if the represents a week day; otherwise <b>false</b>.
        /// </returns>
        /// <remarks>
        /// All days other than Saturday and Sunday are considered week days.
        /// </remarks>
        public static bool IsWeekDay(this DateTime source)
        {
            return source.DayOfWeek != DayOfWeek.Saturday &&
                   source.DayOfWeek != DayOfWeek.Sunday;
        }

        /// <summary>
        /// Returns a new <see cref="DateTime"/> that adds the specified number of week days to a specified value.
        /// </summary>
        /// <param name="source">
        /// The input <see cref="DateTime"/>, which acts as the <b>this</b> instance for the extension method.
        /// </param>
        /// <param name="value">
        /// A number of whole and fractional days. The <i>value</i> parameter can be negative or positive.
        /// </param>
        /// <returns>
        /// An object whose value is the sum of the date and time represented by this instance and the number of week days represented by <i>value</i>.
        /// </returns>
        /// <remarks>
        /// All days other than Saturday and Sunday are considered week days.
        /// </remarks>
        public static DateTime AddWeekDays(this DateTime source, double value)
        {
            // A unit will be +/- 1 day.
            var unit = Math.Sign(value) * 1.0;

            // Start increasing the date by units from the initial date.
            var result = source;

            // When testing for zero, allow a margin for precision error.
            while (!value.IsEquivalentTo(0.0, 0.0001))
            {
                if (Math.Abs(value) < 1.0)
                {
                    // There is less than one full day to add so we need to see whether adding it will take us past midnight.
                    var temp = result.AddDays(value);

                    if (temp.Date == result.Date || temp.IsWeekDay())
                    {
                        // Adding the partial day did not take us into a weekend day so we're done.
                        result = temp;
                        value = 0.0;
                    }
                    else
                    {
                        // Adding the partial day took us into a weekend day so we need to add another day.
                        result = result.AddDays(unit);
                    }
                }
                else
                {
                    // Add a single day.
                    result = result.AddDays(unit);

                    if (result.IsWeekDay())
                    {
                        // Adding a day did not take us into a weekend day so we can reduce the remaining value to add.
                        value -= unit;
                    }
                }
            }

            return result;
        }
    }
}
