﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Wunnell.Extensions
{
    /// <summary>
    /// Contains methods that extend the <see cref="IEnumerable{T}"/> interface.
    /// </summary>
    public static class EnumerableExtensions
    {
        /// <summary>
        /// A random number generator.
        /// </summary>
        private static Random rng;

        /// <summary>
        /// Randomises the items in an enumerable list.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the items in the list.
        /// </typeparam>
        /// <param name="source">
        /// The input list.
        /// </param>
        /// <returns>
        /// The items from the input list in random order.
        /// </returns>
        public static IEnumerable<T> Randomize<T>(this IEnumerable<T> source)
        {
            EnsureRandomInitialized();

            return source.Randomize(rng);
        }

        /// <summary>
        /// Randomises the items in an enumerable list.
        /// </summary>
        /// <typeparam name="T">
        /// The type of the items in the list.
        /// </typeparam>
        /// <param name="source">
        /// The input list.
        /// </param>
        /// <param name="random">
        /// A random number generator.
        /// </param>
        /// <returns>
        /// The items from the input list in random order.
        /// </returns>
        /// <remarks>
        /// Because a <see cref="Random"/> object with a known seed can be used, this overload is good for testing purposes.
        /// </remarks>
        public static IEnumerable<T> Randomize<T>(this IEnumerable<T> source, Random random)
        {
            return source.OrderBy(o => random.NextDouble());
        }

        /// <summary>
        /// Initialises the random number generator if it is not already.
        /// </summary>
        private static void EnsureRandomInitialized()
        {
            rng = rng ?? new Random();
        }
    }
}
