﻿using System;

namespace Wunnell.Extensions
{
    /// <summary>
    /// Contains methods that extend the <see cref="Int32"/> structure.
    /// </summary>
    public static class Int32Extensions
    {
        /// <summary>
        /// Calculates the factorial of a number.
        /// </summary>
        /// <param name="source">
        /// The input number.
        /// </param>
        /// <returns>
        /// 1 if the input is 0; otherwise the product of the series {1, 2, 3, ..., n} where <i>n</i> is the input.
        /// </returns>
        public static int Factorial(this int source)
        {
            if (source < 0)
            {
                throw new ArgumentOutOfRangeException(nameof(source), "Value cannot be negative.");
            }

            if (source == 0)
            {
                // 0! is 1 by definition.
                return 1;
            }

            var product = 1;

            for (var i = 1; i <= source; i++)
            {
                product *= i;
            }

            return product;
        }

        /// <summary>
        /// Calculates the <i>nth</i> triangular number.
        /// </summary>
        /// <param name="source">
        /// The input number.
        /// </param>
        /// <returns>
        /// The sum of the series {0, 1, 2, ..., n} where <i>n</i> is the input.
        /// </returns>
        public static int Triangular(this int source)
        {
            if (source < 0)
            {
                throw new ArgumentOutOfRangeException(nameof(source), "Value cannot be negative.");
            }

            var sum = 0;

            for (var i = 0; i <= source; i++)
            {
                sum += i;
            }

            return sum;
        }

        /// <summary>
        /// Calculates what percentage of a specified value the input is.
        /// </summary>
        /// <param name="source">
        /// The input number.
        /// </param>
        /// <param name="value">
        /// The number to calculate the percentage of.
        /// </param>
        /// <returns>
        /// The percentage of <i>value</i> that the input represents.
        /// </returns>
        public static double AsPercentageOf(this int source, int value)
        {
            return Convert.ToDouble(source).AsPercentageOf(Convert.ToDouble(value));
        }

        /// <summary>
        /// Calculates what percentage of a specified value the input is.
        /// </summary>
        /// <param name="source">
        /// The input number.
        /// </param>
        /// <param name="value">
        /// The number to calculate the percentage of.
        /// </param>
        /// <returns>
        /// The percentage of <i>value</i> that the input represents.
        /// </returns>
        public static double AsPercentageOf(this int source, double value)
        {
            return Convert.ToDouble(source).AsPercentageOf(value);
        }

        /// <summary>
        /// Calculates a proportion of a specified value determined by the input as a percentage.
        /// </summary>
        /// <param name="source">
        /// The input number.
        /// </param>
        /// <param name="value">
        /// The number to calculate the percentage of.
        /// </param>
        /// <returns>
        /// A value that is the percentage of <i>value</i> specified by the input.
        /// </returns>
        public static double PercentOf(this int source, int value)
        {
            return Convert.ToDouble(source).PercentOf(Convert.ToDouble(value));
        }

        /// <summary>
        /// Calculates a proportion of a specified value determined by the input as a percentage.
        /// </summary>
        /// <param name="source">
        /// The input number.
        /// </param>
        /// <param name="value">
        /// The number to calculate the percentage of.
        /// </param>
        /// <returns>
        /// A value that is the percentage of <i>value</i> specified by the input.
        /// </returns>
        public static double PercentOf(this int source, double value)
        {
            return Convert.ToDouble(source).PercentOf(value);
        }
    }
}
