﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Wunnell.Extensions
{
    /// <summary>
    /// Contains methods that extend the <see cref="Random"/> class.
    /// </summary>
    public static class RandomExtensions
    {
        private const int DEFAULT_WEIGHTING = 1;

        /// <summary>
        /// Returns a non-negative random integer that is less than the specified maximum where each value in that range has a weighted probability.
        /// </summary>
        /// <param name="source">
        /// The <see cref="Random"/> object to use to generate the number.
        /// </param>
        /// <param name="maxValue">
        /// The exclusive upper bound of the random number to be generated. maxValue must be greater than or equal to 0.
        /// </param>
        /// <param name="weightings">
        /// The weightings for each of the possible outcomes.
        /// </param>
        /// <returns>
        /// A 32-bit signed integer that is greater than or equal to 0, and less than maxValue; that is, the range of return values ordinarily includes 0 but not maxValue. However, if maxValue equals 0, maxValue is returned.
        /// </returns>
        /// <remarks>
        /// A non-negative weighting must be provided for each possible outcome.  Weightings are a proportion of the total of all weightings.  They are not percentages.
        /// For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
        /// </remarks>
        public static int NextWithWeighting(this Random source,
                                            int maxValue,
                                            int[] weightings)
        {
            return source.NextWithWeighting(0, maxValue, weightings);
        }

        /// <summary>
        /// Returns a random integer that is within a specified range where each value in that range has a weighted probability.
        /// </summary>
        /// <param name="source">
        /// The <see cref="Random"/> object to use to generate the number.
        /// </param>
        /// <param name="minValue">
        /// The inclusive lower bound of the random number returned.
        /// </param>
        /// <param name="maxValue">
        /// The exclusive upper bound of the random number returned. maxValue must be greater than or equal to minValue.
        /// </param>
        /// <param name="weightings">
        /// The weightings for each of the possible outcomes.
        /// </param>
        /// <returns>
        /// A 32-bit signed integer greater than or equal to minValue and less than maxValue; that is, the range of return values includes minValue but not maxValue. If minValue equals maxValue, minValue is returned.
        /// </returns>
        /// <remarks>
        /// A non-negative weighting must be provided for each possible outcome.  Weightings are a proportion of the total of all weightings.  They are not percentages.
        /// For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
        /// </remarks>
        public static int NextWithWeighting(this Random source,
                                            int minValue,
                                            int maxValue,
                                            int[] weightings)
        {
            if (minValue > maxValue)
            {
                throw new ArgumentOutOfRangeException($"'{nameof(minValue)}' cannot be greater than {nameof(maxValue)}.", nameof(minValue));
            }

            if (maxValue > minValue && weightings.Length != maxValue - minValue)
            {
                throw new ArgumentException("A weighting must be provided for all possible outcomes.", nameof(weightings));
            }

            if (weightings.Any(n => n < 0))
            {
                throw new ArgumentException("All weightings must be greater than zero.", nameof(weightings));
            }

            int totalWeightings;

            try
            {
                totalWeightings = weightings.Sum();
            }
            catch (OverflowException e)
            {
                throw new ArgumentException("The sum of all weightings must not be greater than Int32.MaxValue.", e);
            }

            if (totalWeightings == 0)
            {
                throw new ArgumentException("The sum of all weightings must be greater than zero.", nameof(weightings));
            }

            if (minValue == maxValue || minValue == maxValue + 1)
            {
                // There is only one possible value.
                return minValue;
            }

            // Generate a number in the range 0 to 1 less than the total weightings.
            var number = source.Next(totalWeightings);

            var runningWeighting = 0;

            // For each weighting, check whether the number generated falls in that interval.
            for (var i = 0; i < weightings.Length; i++)
            {
                // Sum the weightings so far.
                // E.g. if the weightings are 10, 20, 30 and 40 then the running weighting for each iteration will be:
                // i = 0: runningWeighting = 0 + 10 = 10
                // i = 1: runningWeighting = 10 + 20 = 30
                // i = 2: runningWeighting = 30 + 30 = 60
                // i = 3: runningWeighting = 60 + 40 = 100
                runningWeighting += weightings[i];

                // There is no interval until the running weighting is greater than zero.
                if (runningWeighting > 0 && number < runningWeighting)
                {
                    // The number generated falls within the current weighting interval so get the value from the original range that corresponds to that interval.
                    return minValue + i;
                }
            }

            // If we end up here then something was wrong with the interval and/or the weightings.
            // The validation at the top of the method should ensure that such cases are always caught first.
            throw new Exception("An unexpected error occurred.");
        }

        /// <summary>
        /// Returns a non-negative random integer that is less than the specified maximum where each value in that range has a weighted probability.
        /// </summary>
        /// <param name="source">
        /// The <see cref="Random"/> object to use to generate the number.
        /// </param>
        /// <param name="maxValue">
        /// The exclusive upper bound of the random number to be generated. maxValue must be greater than or equal to 0.
        /// </param>
        /// <param name="weightings">
        /// A dictionary containing the weightings for some or all of the possible outcomes, where the keys are the possible outcomes and the values are the weightings.
        /// </param>
        /// <returns>
        /// A 32-bit signed integer that is greater than or equal to 0, and less than maxValue; that is, the range of return values ordinarily includes 0 but not maxValue. However, if maxValue equals 0, maxValue is returned.
        /// </returns>
        /// <remarks>
        /// A weighting of 1 is used for any values for which a weighting is not explicitly provided.  Weightings are a proportion of the total of all weightings.  They are not percentages.
        /// For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
        /// </remarks>
        public static int NextWithWeighting(this Random source,
                                            int maxValue,
                                            IDictionary<int,int> weightings)
        {
            return source.NextWithWeighting(0, maxValue, weightings);
        }

        /// <summary>
        /// Returns a random integer that is within a specified range where each value in that range has a weighted probability.
        /// </summary>
        /// <param name="source">
        /// The <see cref="Random"/> object to use to generate the number.
        /// </param>
        /// <param name="minValue">
        /// The inclusive lower bound of the random number returned.
        /// </param>
        /// <param name="maxValue">
        /// The exclusive upper bound of the random number returned. maxValue must be greater than or equal to minValue.
        /// </param>
        /// <param name="weightings">
        /// A dictionary containing the weightings for some or all of the possible outcomes, where the keys are the possible outcomes and the values are the weightings.
        /// </param>
        /// <returns>
        /// A 32-bit signed integer greater than or equal to minValue and less than maxValue; that is, the range of return values includes minValue but not maxValue. If minValue equals maxValue, minValue is returned.
        /// </returns>
        /// <remarks>
        /// A weighting of 1 is used for any values for which a weighting is not explicitly provided.  Weightings are a proportion of the total of all weightings.  They are not percentages.
        /// For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
        /// </remarks>
        public static int NextWithWeighting(this Random source,
                                            int minValue,
                                            int maxValue,
                                            IDictionary<int, int> weightings)
        {
            return source.NextWithWeighting(minValue, maxValue, weightings, DEFAULT_WEIGHTING);
        }

        /// <summary>
        /// Returns a non-negative random integer that is less than the specified maximum where each value in that range has a weighted probability.
        /// </summary>
        /// <param name="source">
        /// The <see cref="Random"/> object to use to generate the number.
        /// </param>
        /// <param name="maxValue">
        /// The exclusive upper bound of the random number to be generated. maxValue must be greater than or equal to 0.
        /// </param>
        /// <param name="weightings">
        /// A dictionary containing the weightings for some or all of the possible outcomes, where the keys are the possible outcomes and the values are the weightings.
        /// </param>
        /// <param name="defaultWeighting">
        /// Thw weighting to be used for those values for which one is not explicitly provided.
        /// </param>
        /// <returns>
        /// A 32-bit signed integer that is greater than or equal to 0, and less than maxValue; that is, the range of return values ordinarily includes 0 but not maxValue. However, if maxValue equals 0, maxValue is returned.
        /// </returns>
        /// <remarks>
        /// Weightings are a proportion of the total of all weightings.  They are not percentages.
        /// For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
        /// </remarks>
        public static int NextWithWeighting(this Random source,
                                            int maxValue,
                                            IDictionary<int, int> weightings,
                                            int defaultWeighting)
        {
            return source.NextWithWeighting(0, maxValue, weightings, defaultWeighting);
        }

        /// <summary>
        /// Returns a random integer that is within a specified range where each value in that range has a weighted probability.
        /// </summary>
        /// <param name="source">
        /// The <see cref="Random"/> object to use to generate the number.
        /// </param>
        /// <param name="minValue">
        /// The inclusive lower bound of the random number returned.
        /// </param>
        /// <param name="maxValue">
        /// The exclusive upper bound of the random number returned. maxValue must be greater than or equal to minValue.
        /// </param>
        /// <param name="weightings">
        /// A dictionary containing the weightings for some or all of the possible outcomes, where the keys are the possible outcomes and the values are the weightings.
        /// </param>
        /// <param name="defaultWeighting">
        /// Thw weighting to be used for those values for which one is not explicitly provided.
        /// </param>
        /// <returns>
        /// A 32-bit signed integer greater than or equal to minValue and less than maxValue; that is, the range of return values includes minValue but not maxValue. If minValue equals maxValue, minValue is returned.
        /// </returns>
        /// <remarks>
        /// Weightings are a proportion of the total of all weightings.  They are not percentages.
        /// For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
        /// </remarks>
        public static int NextWithWeighting(this Random source,
                                            int minValue,
                                            int maxValue,
                                            IDictionary<int, int> weightings,
                                            int defaultWeighting)
        {
            if (minValue > maxValue)
            {
                throw new ArgumentOutOfRangeException($"'{nameof(minValue)}' cannot be greater than {nameof(maxValue)}.", nameof(minValue));
            }

            if (weightings.Keys.Any(k => k < minValue ||
                                         minValue == maxValue && k > maxValue ||
                                         minValue < maxValue && k >= maxValue))
            {
                throw new ArgumentException("Weightings cannot be provided for values outside the specified range.");
            }

            if (defaultWeighting < 0)
            {
                throw new ArgumentOutOfRangeException(nameof(defaultWeighting), defaultWeighting, "Value cannot be less than zero.");
            }

            var allWeightings = new List<int>
                                {
                                    weightings.TryGetValue(minValue, out var weighting)
                                        ? weighting
                                        : defaultWeighting
                                };

            for (var i = minValue + 1; i < maxValue; i++)
            {
                allWeightings.Add(weightings.TryGetValue(i, out weighting)
                                      ? weighting
                                      : defaultWeighting);
            }

            return source.NextWithWeighting(minValue, maxValue, allWeightings.ToArray());
        }
    }
}