﻿Imports System.Runtime.CompilerServices

''' <summary>
''' Contains methods that extend the <see cref="Random"/> class.
''' </summary>
Public Module RandomExtensions

    Private Const DEFAULT_WEIGHTING As Integer = 1

    ''' <summary>
    ''' Returns a non-negative random integer that is less than the specified maximum where each value in that range has a weighted probability.
    ''' </summary>
    ''' <param name="source">
    ''' The <see cref="Random"/> object to use to generate the number.
    ''' </param>
    ''' <param name="maxValue">
    ''' The exclusive upper bound of the random number to be generated. maxValue must be greater than or equal to 0.
    ''' </param>
    ''' <param name="weightings">
    ''' The weightings for each of the possible outcomes.
    ''' </param>
    ''' <returns>
    ''' A 32-bit signed integer that is greater than or equal to 0, and less than maxValue; that is, the range of return values ordinarily includes 0 but not maxValue. However, if maxValue equals 0, maxValue is returned.
    ''' </returns>
    ''' <remarks>
    ''' A non-negative weighting must be provided for each possible outcome.  Weightings are a proportion of the total of all weightings.  They are not percentages.
    ''' For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
    ''' </remarks>
    <Extension>
    Public Function NextWithWeighting(source As Random,
                                      maxValue As Integer,
                                      weightings As Integer()) As Integer
        Return source.NextWithWeighting(0, maxValue, weightings)
    End Function

    ''' <summary>
    ''' Returns a random integer that is within a specified range where each value in that range has a weighted probability.
    ''' </summary>
    ''' <param name="source">
    ''' The <see cref="Random"/> object to use to generate the number.
    ''' </param>
    ''' <param name="minValue">
    ''' The inclusive lower bound of the random number returned.
    ''' </param>
    ''' <param name="maxValue">
    ''' The exclusive upper bound of the random number returned. maxValue must be greater than or equal to minValue.
    ''' </param>
    ''' <param name="weightings">
    ''' The weightings for each of the possible outcomes.
    ''' </param>
    ''' <returns>
    ''' A 32-bit signed integer greater than or equal to minValue and less than maxValue; that is, the range of return values includes minValue but not maxValue. If minValue equals maxValue, minValue is returned.
    ''' </returns>
    ''' <remarks>
    ''' A non-negative weighting must be provided for each possible outcome.  Weightings are a proportion of the total of all weightings.  They are not percentages.
    ''' For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
    ''' </remarks>
    <Extension>
    Public Function NextWithWeighting(source As Random,
                                      minValue As Integer,
                                      maxValue As Integer,
                                      weightings As Integer()) As Integer
        If minValue > maxValue Then
            Throw New ArgumentOutOfRangeException($"'{NameOf(minValue)}' cannot be greater than {NameOf(maxValue)}.", NameOf(minValue))
        End If

        If maxValue > minValue AndAlso weightings.Length <> maxValue - minValue Then
            Throw New ArgumentException("A weighting must be provided for all possible outcomes.", NameOf(weightings))
        End If

        If weightings.Any(Function(n) n < 0) Then
            Throw New ArgumentException("All weightings must be greater than zero.", NameOf(weightings))
        End If

        Dim totalWeightings As Integer

        Try
            totalWeightings = weightings.Sum()
        Catch ex As OverflowException
            Throw New ArgumentOutOfRangeException("The sum of all weightings must not be greater than Int32.MaxValue.", ex)
        End Try

        If totalWeightings = 0 Then
            Throw New ArgumentException("The sum of all weightings must be greater than zero.", NameOf(weightings))
        End If

        If minValue = maxValue OrElse minValue = maxValue + 1 Then
            'There is only one possible value.
            Return minValue
        End If

        'Generate a number in the range 0 to 1 less than the total weightings.
        Dim number = source.Next(totalWeightings)

        Dim runningWeighting As Integer

        'For each weighting, check whether the number generated falls in that interval.
        For i = 0 To weightings.GetUpperBound(0)
            'Sum the weightings so far.
            'E.g. if the weightings are 10, 20, 30 and 40 then the running weighting for each iteration will be:
            'i = 0: runningWeighting = 0 + 10 = 10
            'i = 1: runningWeighting = 10 + 20 = 30
            'i = 2: runningWeighting = 30 + 30 = 60
            'i = 3: runningWeighting = 60 + 40 = 100
            runningWeighting += weightings(i)

            'There is no interval until the running weighting is greater than zero.
            If runningWeighting > 0 AndAlso number < runningWeighting Then
                'The number generated falls within the current weighting interval so get the value from the original range that corresponds to that interval.
                Return minValue + i
            End If
        Next

        'If we end up here then something was wrong with the interval and/or the weightings.
        'The validation at the top of the method should ensure that such cases are always caught first.
        Throw New Exception("An unexpected error occurred.")
    End Function

    ''' <summary>
    ''' Returns a non-negative random integer that is less than the specified maximum where each value in that range has a weighted probability.
    ''' </summary>
    ''' <param name="source">
    ''' The <see cref="Random"/> object to use to generate the number.
    ''' </param>
    ''' <param name="maxValue">
    ''' The exclusive upper bound of the random number to be generated. maxValue must be greater than or equal to 0.
    ''' </param>
    ''' <param name="weightings">
    ''' A dictionary containing the weightings for some Or all of the possible outcomes, where the keys are the possible outcomes And the values are the weightings.
    ''' </param>
    ''' <returns>
    ''' A 32-bit signed integer that is greater than or equal to 0, and less than maxValue; that is, the range of return values ordinarily includes 0 but not maxValue. However, if maxValue equals 0, maxValue is returned.
    ''' </returns>
    ''' <remarks>
    ''' A weighting of 1 is used for any values for which a weighting is not explicitly provided.  Weightings are a proportion of the total of all weightings.  They are not percentages.
    ''' For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
    ''' </remarks>
    <Extension>
    Public Function NextWithWeighting(source As Random,
                                      maxValue As Integer,
                                      weightings As IDictionary(Of Integer, Integer)) As Integer
        Return source.NextWithWeighting(0, maxValue, weightings)
    End Function

    ''' <summary>
    ''' Returns a random integer that is within a specified range where each value in that range has a weighted probability.
    ''' </summary>
    ''' <param name="source">
    ''' The <see cref="Random"/> object to use to generate the number.
    ''' </param>
    ''' <param name="minValue">
    ''' The inclusive lower bound of the random number returned.
    ''' </param>
    ''' <param name="maxValue">
    ''' The exclusive upper bound of the random number returned. maxValue must be greater than or equal to minValue.
    ''' </param>
    ''' <param name="weightings">
    ''' A dictionary containing the weightings for some Or all of the possible outcomes, where the keys are the possible outcomes And the values are the weightings.
    ''' </param>
    ''' <returns>
    ''' A 32-bit signed integer greater than or equal to minValue and less than maxValue; that is, the range of return values includes minValue but not maxValue. If minValue equals maxValue, minValue is returned.
    ''' </returns>
    ''' <remarks>
    ''' A weighting of 1 is used for any values for which a weighting is not explicitly provided.  Weightings are a proportion of the total of all weightings.  They are not percentages.
    ''' For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
    ''' </remarks>
    <Extension>
    Public Function NextWithWeighting(source As Random,
                                      minValue As Integer,
                                      maxValue As Integer,
                                      weightings As IDictionary(Of Integer, Integer)) As Integer
        Return source.NextWithWeighting(minValue, maxValue, weightings, DEFAULT_WEIGHTING)
    End Function

    ''' <summary>
    ''' Returns a non-negative random integer that is less than the specified maximum where each value in that range has a weighted probability.
    ''' </summary>
    ''' <param name="source">
    ''' The <see cref="Random"/> object to use to generate the number.
    ''' </param>
    ''' <param name="maxValue">
    ''' The exclusive upper bound of the random number to be generated. maxValue must be greater than or equal to 0.
    ''' </param>
    ''' <param name="weightings">
    ''' A dictionary containing the weightings for some Or all of the possible outcomes, where the keys are the possible outcomes And the values are the weightings.
    ''' </param>
    ''' <param name="defaultWeighting">
    ''' Thw weighting to be used for those values for which one Is Not explicitly provided.
    ''' </param>
    ''' <returns>
    ''' A 32-bit signed integer that is greater than or equal to 0, and less than maxValue; that is, the range of return values ordinarily includes 0 but not maxValue. However, if maxValue equals 0, maxValue is returned.
    ''' </returns>
    ''' <remarks>
    ''' Weightings are a proportion of the total of all weightings.  They are not percentages.
    ''' For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
    ''' </remarks>
    <Extension>
    Public Function NextWithWeighting(source As Random,
                                      maxValue As Integer,
                                      weightings As IDictionary(Of Integer, Integer),
                                      defaultWeighting As Integer) As Integer
        Return source.NextWithWeighting(0, maxValue, weightings, defaultWeighting)
    End Function

    ''' <summary>
    ''' Returns a random integer that is within a specified range where each value in that range has a weighted probability.
    ''' </summary>
    ''' <param name="source">
    ''' The <see cref="Random"/> object to use to generate the number.
    ''' </param>
    ''' <param name="minValue">
    ''' The inclusive lower bound of the random number returned.
    ''' </param>
    ''' <param name="maxValue">
    ''' The exclusive upper bound of the random number returned. maxValue must be greater than or equal to minValue.
    ''' </param>
    ''' <param name="weightings">
    ''' A dictionary containing the weightings for some Or all of the possible outcomes, where the keys are the possible outcomes And the values are the weightings.
    ''' </param>
    ''' <param name="defaultWeighting">
    ''' Thw weighting to be used for those values for which one Is Not explicitly provided.
    ''' </param>
    ''' <returns>
    ''' A 32-bit signed integer greater than or equal to minValue and less than maxValue; that is, the range of return values includes minValue but not maxValue. If minValue equals maxValue, minValue is returned.
    ''' </returns>
    ''' <remarks>
    ''' Weightings are a proportion of the total of all weightings.  They are not percentages.
    ''' For instance, if there are three possible outcomes and the weightings are 1, 2 and 3 then the first outcome will result in about 1/6 of the time, the second outcome will result about 1/3 of the time and the third outcome will result about 1/2 of the time.
    ''' </remarks>
    <Extension>
    Public Function NextWithWeighting(source As Random,
                                      minValue As Integer,
                                      maxValue As Integer,
                                      weightings As IDictionary(Of Integer, Integer),
                                      defaultWeighting As Integer) As Integer
        If minValue > maxValue Then
            Throw New ArgumentOutOfRangeException($"'{NameOf(minValue)}' cannot be greater than {NameOf(maxValue)}.", NameOf(minValue))
        End If

        If weightings.Keys.Any(Function(k) k < minValue OrElse
                                           minValue = maxValue AndAlso k > maxValue OrElse
                                           minValue < maxValue AndAlso k >= maxValue) Then
            Throw New ArgumentException("Weightings cannot be provided for values outside the specified range.")
        End If

        If defaultWeighting < 0 Then
            Throw New ArgumentOutOfRangeException(NameOf(defaultWeighting), defaultWeighting, "Value cannot be less than zero.")
        End If

        Dim weighting As Integer
        Dim allWeightings As New List(Of Integer) From {If(weightings.TryGetValue(minValue, weighting),
                                                           weighting,
                                                           defaultWeighting)}

        For i = minValue + 1 To maxValue - 1 Step 1
            allWeightings.Add(If(weightings.TryGetValue(i, weighting),
                                 weighting,
                                 defaultWeighting))
        Next

        Return source.NextWithWeighting(minValue, maxValue, allWeightings.ToArray())
    End Function

End Module
