﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Numerics;
using System.Threading;

namespace Wunnell.Extensions
{
    /// <summary>
    /// Contains methods that extend the <see cref="String"/> class.
    /// </summary>
    public static class StringExtensions
    {
        #region Lines

        /// <summary>
        /// Splits a string into lines.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// An array whose elements contain the lines from this instance.
        /// </returns>
        /// <remarks>
        /// Line breaks are considered to be carriage-return/line-feed pairs, single carriage-returns and single line-feeds.
        /// </remarks>
        public static string[] ToLines(this string source)
        {
            return source.ToLines(StringSplitOptions.None);
        }

        /// <summary>
        /// Splits a string into lines. You can specify whether empty lines should be included.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="options">
        /// <see cref="StringSplitOptions">StringSplitOptions.RemoveEmptyEntries</see> to omit empty array elements from the array returned; or <see cref="StringSplitOptions">StringSplitOptions.None</see> to include empty array elements in the array returned.
        /// </param>
        /// <returns>
        /// An array whose elements contain the lines from this instance.
        /// </returns>
        /// <remarks>
        /// Line breaks are considered to be carriage-return/line-feed pairs, single carriage-returns and single line-feeds.
        /// </remarks>
        public static string[] ToLines(this string source, StringSplitOptions options)
        {
            return source.Split(new[] { "\r\n", "\r", "\n" },
                                options);
        }

        /// <summary>
        /// Splits a string into a maximum number of lines. You can specify whether empty lines should be included.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="count">
        /// The maximum number of lines to return.
        /// </param>
        /// <param name="options">
        /// <see cref="StringSplitOptions">StringSplitOptions.RemoveEmptyEntries</see> to omit empty array elements from the array returned; or <see cref="StringSplitOptions">StringSplitOptions.None</see> to include empty array elements in the array returned.
        /// </param>
        /// <returns>
        /// An array whose elements contain the lines from this instance.
        /// </returns>
        /// <remarks>
        /// Line breaks are considered to be carriage-return/line-feed pairs, single carriage-returns and single line-feeds.
        /// </remarks>
        public static string[] ToLines(this string source, int count, StringSplitOptions options)
        {
            return source.Split(new[] {"\r\n", "\r", "\n"},
                                count,
                                options);
        }

        #endregion Lines

        #region Convert to Title Case

        /// <summary>
        /// Returns a copy of a string converted to title case (except for words that are entirely in uppercase, which are considered to be acronyms).
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The title case equivalent of the input string.
        /// </returns>
        public static string ToTitle(this string source)
        {
            return source.ToTitle(CultureInfo.CurrentCulture);
        }

        /// <summary>
        /// Returns a copy of a string converted to title case.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="convertAcronyms">
        /// <b>true</b> to convert words in all upper-case; otherwise, <b>false</b>.
        /// </param>
        /// <returns>
        /// The title case equivalent of the input string.
        /// </returns>
        public static string ToTitle(this string source, bool convertAcronyms)
        {
            return source.ToTitle(CultureInfo.CurrentCulture, convertAcronyms);
        }

        /// <summary>
        /// Returns a copy of a string converted to title case, using the casing rules of the specified culture (except for words that are entirely in uppercase, which are considered to be acronyms).
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="culture">
        /// An object that supplies culture-specific casing rules.
        /// </param>
        /// <returns>
        /// The title case equivalent of the input string.
        /// </returns>
        public static string ToTitle(this string source, CultureInfo culture)
        {
            return source.ToTitle(culture, false);
        }

        /// <summary>
        /// Returns a copy of a string converted to title case, using the casing rules of the specified culture.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="culture">
        /// An object that supplies culture-specific casing rules.
        /// </param>
        /// <param name="convertAcronyms">
        /// <b>true</b> to convert words in all upper-case; otherwise, <b>false</b>.
        /// </param>
        /// <returns>
        /// The title case equivalent of the input string.
        /// </returns>
        public static string ToTitle(this string source, CultureInfo culture, bool convertAcronyms)
        {
            return culture.TextInfo.ToTitleCase(convertAcronyms
                                                    ? source.ToLower(culture)
                                                    : source);
        }

        /// <summary>
        /// Returns a copy of a string converted to title case, using the casing rules of the invariant culture (except for words that are entirely in uppercase, which are considered to be acronyms).
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The title case equivalent of the input string.
        /// </returns>
        public static string ToTitleInvariant(this string source)
        {
            return source.ToTitle(CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Returns a copy of a string converted to title case, using the casing rules of the invariant culture.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="convertAcronyms">
        /// <b>true</b> to convert words in all upper-case; otherwise, <b>false</b>.
        /// </param>
        /// <returns>
        /// The title case equivalent of the input string.
        /// </returns>
        public static string ToTitleInvariant(this string source, bool convertAcronyms)
        {
            return source.ToTitle(CultureInfo.InvariantCulture, convertAcronyms);
        }

        #endregion Convert to Title Case

        #region Substring

        /// <summary>
        /// Retrieves a substring of up to the specified length from the beginning of a string instance.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="length">
        /// The maximum number of characters to include in the substring.
        /// If <i>length</i> is greater than the length of the input string, the entire input is returned.
        /// </param>
        /// <returns>
        /// A string containing the specified number of characters from the beginning of the input string if it is at least that long; otherwise, the entire input string.
        /// </returns>
        public static string LeftSubstring(this string source, int length)
        {
            return source.MidSubstring(0, length);
        }

        /// <summary>
        /// Retrieves a substring of up to the specified length from the specified index of a string instance.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="startIndex">
        /// The index of the first character in the input to be included in the output.
        /// If <i>startIndex</i> is greater than the index of the last character in the input string, an empty string is returned.
        /// </param>
        /// <param name="length">
        /// The maximum number of characters to include in the substring.
        /// </param>
        /// <returns>
        /// A string containing the specified number of characters from the specified index of the input string if it is at least that long; otherwise, only the characters from the input string that fall within those bounds.
        /// </returns>
        public static string MidSubstring(this string source, int startIndex, int length)
        {
            startIndex = Math.Min(Math.Max(startIndex, 0), source.Length);
            length = Math.Min(Math.Max(length, 0), source.Length - startIndex);

            return source.Substring(startIndex, length);
        }

        /// <summary>
        /// Retrieves a substring of up to the specified length from the end of a string instance.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="length">
        /// The maximum number of characters to include in the substring.
        /// If <i>length</i> is greater than the length of the input string, the entire input is returned.
        /// </param>
        /// <returns>
        /// A string containing the specified number of characters from the end of the input string if it is at least that long; otherwise, the entire input string.
        /// </returns>
        public static string RightSubstring(this string source, int length)
        {
            return source.MidSubstring(source.Length - length, length);
        }

        /// <summary>
        /// Splits a string instance into substrings of the specified length.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="length">
        /// The maximum number of characters to include in each substring.
        /// If the length of the input string is not a multiple of <i>length</i>, the last substring in the output will contain fewer characters.
        /// </param>
        /// <returns>
        /// An array of substrings from the input string.  If the input string length is less than <i>length</i>, the array will contain just the input string.
        /// Each element of the array will contain exactly <i>length</i> characters except the last, which will contain fewer characters if the input string is not a multiple of <i>length</i>.
        /// </returns>
        public static string[] SubstringsByLength(this string source, int length)
        {
            if (source == null)
            {
                throw new ArgumentNullException(nameof(source));
            }

            if (length <= 0)
            {
                throw new ArgumentException($"{length} is not a valid value for '{nameof(length)}'. Value must be greater than zero.", nameof(length));
            }

            var sourceLength = source.Length;

            if (sourceLength <= length)
            {
                // Return an array containing just the original string.
                return new[] {source};
            }

            var substrings = new List<string> {source.LeftSubstring(length)};
            var index = length;

            while (index < sourceLength)
            {
                substrings.Add(source.MidSubstring(index, length));
                index += length;
            }

            return substrings.ToArray();
        }

        /// <summary>
        /// Splits a string instance into the specified number of substrings.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="count">
        /// The maximum number of substrings to include in the output.
        /// If <i>count</i> is greater than the length of the input string, the output will contain all the individual characters from the input.
        /// </param>
        /// <returns>
        /// An array of substrings from the input string.  If the inout string length is less than <i>count</i>, the array will as many elements as there are characters in the inout string.
        /// The length of the substrings will vary by no more than one character, with the elements at the beginning being longer than those at the end if the input string length is not a multiple of <i>count</i>.
        /// </returns>
        public static string[] SubstringsByCount(this string source, int count)
        {
            if (source == null)
            {
                throw new ArgumentNullException(nameof(source));
            }

            if (count <= 0)
            {
                throw new ArgumentException($"{count} is not a valid value for '{nameof(count)}'. Value must be greater than zero.", nameof(count));
            }

            if (count == 1)
            {
                // Return an array containing just the original string.
                return new[] { source };
            }

            var sourceLength = source.Length;

            if (sourceLength <= count)
            {
                // Return an array containing each character as a string.
                return Array.ConvertAll(source.ToCharArray(), ch => ch.ToString());
            }

            var substrings = new string[count];
            var minLength = Math.DivRem(sourceLength, count, out var overflow);
            var startIndex = 0;

            for (var i = 0; i < count; i++)
            {
                var length = i < overflow
                                 ? minLength + 1
                                 : minLength;

                substrings[i] = source.Substring(startIndex, length);
                startIndex += length;
            }

            return substrings;
        }

        #endregion Substring

        #region Contains

        /// <summary>
        /// Returns a value indicating whether a specified Unicode character occurs within the inpout string.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="value">
        /// A Unicode character to seek.
        /// </param>
        /// <returns>
        /// <b>true</b> if <i>value</i> is found within the input string; otherwise, <b>false</b>.
        /// </returns>
        public static bool Contains(this string source, char value)
        {
            return source.IndexOf(value) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether a specified Unicode character occurs within the inpout string. The search starts at a specified character position.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="value">
        /// A Unicode character to seek.
        /// </param>
        /// <param name="startIndex">
        /// The search starting position.
        /// </param>
        /// <returns>
        /// <b>true</b> if <i>value</i> is found within the input string; otherwise, <b>false</b>.
        /// </returns>
        public static bool Contains(this string source,
                                    char value,
                                    int startIndex)
        {
            return source.IndexOf(value, startIndex) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether a specified substring occurs within the inpout string. The search starts at a specified character position.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="value">
        /// The string to seek.
        /// </param>
        /// <param name="startIndex">
        /// The search starting position.
        /// </param>
        /// <returns>
        /// <b>true</b> if <i>value</i> is found within the input string; otherwise, <b>false</b>. If <i>value</i> is <see cref="string.Empty"/>, the return value is <b>true</b>.
        /// </returns>
        public static bool Contains(this string source,
                                    string value,
                                    int startIndex)
        {
            return source.IndexOf(value, startIndex) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether a specified substring occurs within the inpout string.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="value">
        /// The string to seek.
        /// </param>
        /// <param name="comparisonType">
        /// One of the enumeration values that specifies the rules for the search.
        /// </param>
        /// <returns>
        /// <b>true</b> if <i>value</i> is found within the input string; otherwise, <b>false</b>. If <i>value</i> is <see cref="string.Empty"/>, the return value is <b>true</b>.
        /// </returns>
        public static bool Contains(this string source,
                                    string value,
                                    StringComparison comparisonType)
        {
            return source.IndexOf(value, comparisonType) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether a specified Unicode character occurs within the inpout string.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="value">
        /// A Unicode character to seek.
        /// </param>
        /// <param name="startIndex">
        /// The search starting position.
        /// </param>
        /// <param name="count">
        /// The number of character positions to examine.
        /// </param>
        /// <returns>
        /// <b>true</b> if <i>value</i> is found within the input string; otherwise, <b>false</b>.
        /// </returns>
        public static bool Contains(this string source,
                                    char value,
                                    int startIndex,
                                    int count)
        {
            return source.IndexOf(value, startIndex, count) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether a specified substring occurs within the inpout string.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="value">
        /// The string to seek.
        /// </param>
        /// <param name="startIndex">
        /// The search starting position.
        /// </param>
        /// <param name="count">
        /// The number of character positions to examine.
        /// </param>
        /// <returns>
        /// <b>true</b> if <i>value</i> is found within the input string; otherwise, <b>false</b>. If <i>value</i> is <see cref="string.Empty"/>, the return value is <b>true</b>.
        /// </returns>
        public static bool Contains(this string source,
                                    string value,
                                    int startIndex,
                                    int count)
        {
            return source.IndexOf(value, startIndex, count) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether a specified substring occurs within the inpout string.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="value">
        /// The string to seek.
        /// </param>
        /// <param name="startIndex">
        /// The search starting position.
        /// </param>
        /// <param name="comparisonType">
        /// One of the enumeration values that specifies the rules for the search.
        /// </param>
        /// <returns>
        /// <b>true</b> if <i>value</i> is found within the input string; otherwise, <b>false</b>. If <i>value</i> is <see cref="string.Empty"/>, the return value is <b>true</b>.
        /// </returns>
        public static bool Contains(this string source,
                                    string value,
                                    int startIndex,
                                    StringComparison comparisonType)
        {
            return source.IndexOf(value, startIndex, comparisonType) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether a specified substring occurs within the inpout string.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="value">
        /// The string to seek.
        /// </param>
        /// <param name="startIndex">
        /// The search starting position.
        /// </param>
        /// <param name="count">
        /// The number of character positions to examine.
        /// </param>
        /// <param name="comparisonType">
        /// One of the enumeration values that specifies the rules for the search.
        /// </param>
        /// <returns>
        /// <b>true</b> if <i>value</i> is found within the input string; otherwise, <b>false</b>. If <i>value</i> is <see cref="string.Empty"/>, the return value is <b>true</b>.
        /// </returns>
        public static bool Contains(this string source,
                                    string value,
                                    int startIndex,
                                    int count,
                                    StringComparison comparisonType)
        {
            return source.IndexOf(value, startIndex, count, comparisonType) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether any character in a specified array of Unicode characters occur within the inpout string.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="anyOf">
        /// A Unicode character array containing one or more characters to seek.
        /// </param>
        /// <returns>
        /// <b>true</b> if any character in <i>anyOf</i> is found within the input string; otherwise, <b>false</b>.
        /// </returns>
        public static bool ContainsAny(this string source, char[] anyOf)
        {
            return source.IndexOfAny(anyOf) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether any character in a specified array of Unicode characters occur within the inpout string.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="anyOf">
        /// A Unicode character array containing one or more characters to seek.
        /// </param>
        /// <param name="startIndex">
        /// The search starting position.
        /// </param>
        /// <returns>
        /// <b>true</b> if any character in <i>anyOf</i> is found within the input string; otherwise, <b>false</b>.
        /// </returns>
        public static bool ContainsAny(this string source,
                                       char[] anyOf,
                                       int startIndex)
        {
            return source.IndexOfAny(anyOf, startIndex) != -1;
        }

        /// <summary>
        /// Returns a value indicating whether any character in a specified array of Unicode characters occur within the inpout string.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <param name="anyOf">
        /// A Unicode character array containing one or more characters to seek.
        /// </param>
        /// <param name="startIndex">
        /// The search starting position.
        /// </param>
        /// <param name="count">
        /// The number of character positions to examine.
        /// </param>
        /// <returns>
        /// <b>true</b> if any character in <i>anyOf</i> is found within the input string; otherwise, <b>false</b>.
        /// </returns>
        public static bool ContainsAny(this string source,
                                       char[] anyOf,
                                       int startIndex,
                                       int count)
        {
            return source.IndexOfAny(anyOf, startIndex, count) != -1;
        }

        #endregion Contains

        #region Convert to Number by Base

        /// <summary>
        /// Converts a string containing binary digits in a <see cref="byte"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the binary string.
        /// </returns>
        public static byte ToByteFromBinary(this string source)
        {
            return Convert.ToByte(source, 2);
        }

        /// <summary>
        /// Converts a string containing binary digits in a <see cref="short"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the binary string.
        /// </returns>
        public static short ToInt16FromBinary(this string source)
        {
            return Convert.ToInt16(source, 2);
        }

        /// <summary>
        /// Converts a string containing binary digits in a <see cref="int"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the binary string.
        /// </returns>
        public static int ToInt32FromBinary(this string source)
        {
            return Convert.ToInt32(source, 2);
        }

        /// <summary>
        /// Converts a string containing binary digits in a <see cref="long"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the binary string.
        /// </returns>
        public static long ToInt64FromBinary(this string source)
        {
            return Convert.ToInt64(source, 2);
        }

        /// <summary>
        /// Converts a string containing octal digits in a <see cref="byte"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the octal string.
        /// </returns>
        public static byte ToByteFromOctal(this string source)
        {
            return Convert.ToByte(source, 8);
        }

        /// <summary>
        /// Converts a string containing octal digits in a <see cref="short"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the octal string.
        /// </returns>
        public static short ToInt16FromOctal(this string source)
        {
            return Convert.ToInt16(source, 8);
        }

        /// <summary>
        /// Converts a string containing octal digits in a <see cref="int"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the octal string.
        /// </returns>
        public static int ToInt32FromOctal(this string source)
        {
            return Convert.ToInt32(source, 8);
        }

        /// <summary>
        /// Converts a string containing octal digits in a <see cref="long"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the octal string.
        /// </returns>
        public static long ToInt64FromOctal(this string source)
        {
            return Convert.ToInt64(source, 8);
        }

        /// <summary>
        /// Converts a string containing decimal digits in a <see cref="byte"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the decimal string.
        /// </returns>
        public static byte ToByteFromDecimal(this string source)
        {
            return Convert.ToByte(source, 10);
        }

        /// <summary>
        /// Converts a string containing decimal digits in a <see cref="short"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the decimal string.
        /// </returns>
        public static short ToInt16FromDecimal(this string source)
        {
            return Convert.ToInt16(source, 10);
        }

        /// <summary>
        /// Converts a string containing decimal digits in a <see cref="int"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the decimal string.
        /// </returns>
        public static int ToInt32FromDecimal(this string source)
        {
            return Convert.ToInt32(source, 10);
        }

        /// <summary>
        /// Converts a string containing decimal digits in a <see cref="long"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the decimal string.
        /// </returns>
        public static long ToInt64FromDecimal(this string source)
        {
            return Convert.ToInt64(source, 10);
        }

        /// <summary>
        /// Converts a string containing hexadecimal digits in a <see cref="byte"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the hexadecimal string.
        /// </returns>
        public static byte ToByteFromHexadecimal(this string source)
        {
            return Convert.ToByte(source, 16);
        }

        /// <summary>
        /// Converts a string containing hexadecimal digits in a <see cref="short"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the hexadecimal string.
        /// </returns>
        public static short ToInt16FromHexadecimal(this string source)
        {
            return Convert.ToInt16(source, 16);
        }

        /// <summary>
        /// Converts a string containing hexadecimal digits in a <see cref="int"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the hexadecimal string.
        /// </returns>
        public static int ToInt32FromHexadecimal(this string source)
        {
            return Convert.ToInt32(source, 16);
        }

        /// <summary>
        /// Converts a string containing hexadecimal digits in a <see cref="long"/>.
        /// </summary>
        /// <param name="source">
        /// The input string.
        /// </param>
        /// <returns>
        /// The number represented by the hexadecimal string.
        /// </returns>
        public static long ToInt64FromHexadecimal(this string source)
        {
            return Convert.ToInt64(source, 16);
        }

        #endregion Convert to Number by Base

        #region Numeric Validation

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="byte"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="byte"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsByte(this string source)
        {
            return byte.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="sbyte"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="sbyte"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsSByte(this string source)
        {
            return sbyte.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="short"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="short"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsInt16(this string source)
        {
            return short.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="ushort"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="ushort"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsUInt16(this string source)
        {
            return ushort.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="int"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="int"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsInt32(this string source)
        {
            return int.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="uint"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="uint"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsUInt32(this string source)
        {
            return uint.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="long"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="long"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsInt64(this string source)
        {
            return long.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="ulong"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="ulong"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsUInt64(this string source)
        {
            return ulong.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="float"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="float"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsSingle(this string source)
        {
            return float.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="double"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="double"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsDouble(this string source)
        {
            return double.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="decimal"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="decimal"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsDecimal(this string source)
        {
            return decimal.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="Guid"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="Guid"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsGuid(this string source)
        {
            return Guid.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="Guid"/> from the specified format.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <param name="format">
        /// One of the following specifiers that indicates the exact format to use when interpreting input: "N", "D", "B", "P", or "X".
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="Guid"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsGuid(this string source, string format)
        {
            return Guid.TryParseExact(source, format, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="BigInteger"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="BigInteger"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsBigInteger(this string source)
        {
            return BigInteger.TryParse(source, out _);
        }

        #endregion Numeric Validation

        #region Date/Time Validation

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="DateTime"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="DateTime"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsDateTime(this string source)
        {
            return DateTime.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="DateTime"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <param name="format">
        /// The required format of the input string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="DateTime"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsDateTime(this string source, string format)
        {
            return DateTime.TryParseExact(source, format, null, DateTimeStyles.None, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="DateTime"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <param name="formats">
        /// An array of the allowable formats of the input string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="DateTime"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsDateTime(this string source, params string[] formats)
        {
            return DateTime.TryParseExact(source, formats, null, DateTimeStyles.None, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="TimeSpan"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="TimeSpan"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsTimeSpan(this string source)
        {
            return TimeSpan.TryParse(source, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="TimeSpan"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <param name="format">
        /// The required format of the input string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="TimeSpan"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsTimeSpan(this string source, string format)
        {
            return TimeSpan.TryParseExact(source, format, null, TimeSpanStyles.None, out _);
        }

        /// <summary>
        /// Determines whether a string can be converted to a <see cref="TimeSpan"/>.
        /// </summary>
        /// <param name="source">
        /// The inout string.
        /// </param>
        /// <param name="formats">
        /// An array of the allowable formats of the input string.
        /// </param>
        /// <returns>
        /// <b>true</b> if the inout string can be converted to a <see cref="TimeSpan"/>; otherwise, <b>false</b>.
        /// </returns>
        public static bool IsTimeSpan(this string source, params string[] formats)
        {
            return TimeSpan.TryParseExact(source, formats, null, TimeSpanStyles.None, out _);
        }

        #endregion Date/Time Validation
    }
}