﻿Imports System.Globalization
Imports System.Numerics
Imports System.Runtime.CompilerServices

''' <summary>
''' Contains methods that extend the <see cref="String"/> class.
''' </summary>
Public Module StringExtensions

#Region "Lines"

    ''' <summary>
    ''' Splits a string into lines.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' An array whose elements contain the lines from this instance.
    ''' </returns>
    ''' <remarks>
    ''' Line breaks are considered to be carriage-return/line-feed pairs, single carriage-returns and single line-feeds.
    ''' </remarks>
    <Extension>
    Public Function ToLines(source As String) As String()
        Return source.ToLines(StringSplitOptions.None)
    End Function

    ''' <summary>
    ''' Splits a string into lines. You can specify whether empty lines should be included.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="options">
    ''' <see cref="StringSplitOptions">StringSplitOptions.RemoveEmptyEntries</see> to omit empty array elements from the array returned; or <see cref="StringSplitOptions">StringSplitOptions.None</see> to include empty array elements in the array returned.
    ''' </param>
    ''' <returns>
    ''' An array whose elements contain the lines from this instance.
    ''' </returns>
    ''' <remarks>
    ''' Line breaks are considered to be carriage-return/line-feed pairs, single carriage-returns and single line-feeds.
    ''' </remarks>
    <Extension>
    Public Function ToLines(source As String, options As StringSplitOptions) As String()
        Return source.Split({ControlChars.CrLf, ControlChars.Cr, ControlChars.Lf},
                            options)
    End Function

    ''' <summary>
    ''' Splits a string into a maximum number of lines. You can specify whether empty lines should be included.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="count">
    ''' The maximum number of lines to return.
    ''' </param>
    ''' <param name="options">
    ''' <see cref="StringSplitOptions">StringSplitOptions.RemoveEmptyEntries</see> to omit empty array elements from the array returned; or <see cref="StringSplitOptions">StringSplitOptions.None</see> to include empty array elements in the array returned.
    ''' </param>
    ''' <returns>
    ''' An array whose elements contain the lines from this instance.
    ''' </returns>
    ''' <remarks>
    ''' Line breaks are considered to be carriage-return/line-feed pairs, single carriage-returns and single line-feeds.
    ''' </remarks>
    <Extension>
    Public Function ToLines(source As String, count As Integer, options As StringSplitOptions) As String()
        Return source.Split({ControlChars.CrLf, ControlChars.Cr, ControlChars.Lf},
                            count,
                            options)
    End Function

#End Region 'Lines

#Region "Convert to Title Case"

    ''' <summary>
    ''' Returns a copy of a string converted to title case (except for words that are entirely in uppercase, which are considered to be acronyms).
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The title case equivalent of the input string.
    ''' </returns>
    <Extension>
    Public Function ToTitle(source As String) As String
        Return source.ToTitle(CultureInfo.CurrentCulture)
    End Function

    ''' <summary>
    ''' Returns a copy of a string converted to title case.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="convertAcronyms">
    ''' <b>true</b> to convert words in all upper-case; otherwise, <b>false</b>.
    ''' </param>
    ''' <returns>
    ''' The title case equivalent of the input string.
    ''' </returns>
    <Extension>
    Public Function ToTitle(source As String, convertAcronyms As Boolean) As String
        Return source.ToTitle(CultureInfo.CurrentCulture, convertAcronyms)
    End Function

    ''' <summary>
    ''' Returns a copy of a string converted to title case, using the casing rules of the specified culture (except for words that are entirely in uppercase, which are considered to be acronyms).
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="culture">
    ''' An object that supplies culture-specific casing rules.
    ''' </param>
    ''' <returns>
    ''' The title case equivalent of the input string.
    ''' </returns>
    <Extension>
    Public Function ToTitle(source As String, culture As CultureInfo) As String
        Return source.ToTitle(culture, False)
    End Function

    ''' <summary>
    ''' Returns a copy of a string converted to title case, using the casing rules of the specified culture.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="culture">
    ''' An object that supplies culture-specific casing rules.
    ''' </param>
    ''' <param name="convertAcronyms">
    ''' <b>true</b> to convert words in all upper-case; otherwise, <b>false</b>.
    ''' </param>
    ''' <returns>
    ''' The title case equivalent of the input string.
    ''' </returns>
    <Extension>
    Public Function ToTitle(source As String, culture As CultureInfo, convertAcronyms As Boolean) As String
        Return culture.TextInfo.ToTitleCase(If(convertAcronyms,
                                               source.ToLower(culture),
                                               source))
    End Function

    ''' <summary>
    ''' Returns a copy of a string converted to title case, using the casing rules of the invariant culture (except for words that are entirely in uppercase, which are considered to be acronyms).
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The title case equivalent of the input string.
    ''' </returns>
    <Extension>
    Public Function ToTitleInvariant(source As String) As String
        Return source.ToTitle(CultureInfo.InvariantCulture)
    End Function

    ''' <summary>
    ''' Returns a copy of a string converted to title case, using the casing rules of the invariant culture.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="convertAcronyms">
    ''' <b>true</b> to convert words in all upper-case; otherwise, <b>false</b>.
    ''' </param>
    ''' <returns>
    ''' The title case equivalent of the input string.
    ''' </returns>
    <Extension>
    Public Function ToTitleInvariant(source As String, convertAcronyms As Boolean) As String
        Return source.ToTitle(CultureInfo.InvariantCulture, convertAcronyms)
    End Function

#End Region 'Convert to Title Case

#Region "Substring"

    ''' <summary>
    ''' Retrieves a substring of up to the specified length from the beginning of a string instance.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="length">
    ''' The maximum number of characters to include in the substring.
    ''' If <i>length</i> is greater than the length of the input string, the entire input is returned.
    ''' </param>
    ''' <returns>
    ''' A string containing the specified number of characters from the beginning of the input string if it is at least that long; otherwise, the entire input string.
    ''' </returns>
    <Extension>
    Public Function LeftSubstring(source As String, length As Integer) As String
        Return source.MidSubstring(0, length)
    End Function

    ''' <summary>
    ''' Retrieves a substring of up to the specified length from the specified index of a string instance.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="startIndex">
    ''' The index of the first character in the input to be included in the output.
    ''' If <i>startIndex</i> Is greater than the index of the last character in the input string, an empty string Is returned.
    ''' </param>
    ''' <param name="length">
    ''' The maximum number of characters to include in the substring.
    ''' </param>
    ''' <returns>
    ''' A string containing the specified number of characters from the specified index of the input string if it Is at least that long; otherwise, only the characters from the input string that fall within those bounds.
    ''' </returns>
    <Extension>
    Public Function MidSubstring(source As String,
                                 startIndex As Integer,
                                 length As Integer) As String
        startIndex = Math.Min(Math.Max(startIndex, 0), source.Length)
        length = Math.Min(Math.Max(length, 0), source.Length - startIndex)

        Return source.Substring(startIndex, length)
    End Function

    ''' <summary>
    ''' Retrieves a substring of up to the specified length from the end of a string instance.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="length">
    ''' The maximum number of characters to include in the substring.
    ''' If <i>length</i> Is greater than the length of the input string, the entire input Is returned.
    ''' </param>
    ''' <returns>
    ''' A string containing the specified number of characters from the end of the input string if it Is at least that long; otherwise, the entire input string.
    ''' </returns>
    <Extension>
    Public Function RightSubstring(source As String, length As Integer) As String
        Return source.MidSubstring(source.Length - length, length)
    End Function

    ''' <summary>
    ''' Splits a string instance into substrings of the specified length.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="length">
    ''' The maximum number of characters to include in each substring.
    ''' If the length of the input string Is Not a multiple of <i>length</i>, the last substring in the output will contain fewer characters.
    ''' </param>
    ''' <returns>
    ''' An array of substrings from the input string.  If the input string length Is less than <i>length</i>, the array will contain just the input string.
    ''' Each element of the array will contain exactly <i>length</i> characters except the last, which will contain fewer characters if the input string Is Not a multiple of <i>length</i>.
    ''' </returns>
    <Extension>
    Public Function SubstringsByLength(source As String, length As Integer) As String()
        If source Is Nothing Then
            Throw New ArgumentNullException(NameOf(source))
        End If

        If length <= 0 Then
            Throw New ArgumentException($"{length} is not a valid value for '{NameOf(length)}'. Value must be greater than zero.", NameOf(length))
        End If

        Dim sourceLength = source.Length

        If sourceLength <= length Then
            'Return an array containing just the original string.
            Return {source}
        End If

        Dim substrings As New List(Of String) From {source.LeftSubstring(length)}
        Dim index = length

        While index < sourceLength
            substrings.Add(source.MidSubstring(index, length))
            index += length
        End While

        Return substrings.ToArray()
    End Function

    ''' <summary>
    ''' Splits a string instance into the specified number of substrings.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="count">
    ''' The maximum number of substrings to include in the output.
    ''' If <i>count</i> Is greater than the length of the input string, the output will contain all the individual characters from the input.
    ''' </param>
    ''' <returns>
    ''' An array of substrings from the input string.  If the inout string length Is less than <i>count</i>, the array will as many elements as there are characters in the inout string.
    ''' The length of the substrings will vary by no more than one character, with the elements at the beginning being longer than those at the end if the input string length Is Not a multiple of <i>count</i>.
    ''' </returns>
    <Extension>
    Public Function SubstringsByCount(source As String, count As Integer) As String()
        If source Is Nothing Then
            Throw New ArgumentNullException(NameOf(source))
        End If

        If count <= 0 Then
            Throw New ArgumentException($"{count} is not a valid value for '{NameOf(count)}'. Value must be greater than zero.", NameOf(count))
        End If

        If count = 1 Then
            'Return an array containing just the original string.
            Return {source}
        End If

        Dim sourceLength = source.Length

        If sourceLength <= count Then
            'Return an array containing each character as a string.
            Return Array.ConvertAll(source.ToCharArray(), Function(ch) ch.ToString())
        End If

        Dim substrings(count - 1) As String
        Dim overflow As Integer
        Dim minLength = Math.DivRem(sourceLength, count, overflow)
        Dim startIndex = 0

        For i = 0 To count - 1
            Dim length = If(i < overflow,
                            minLength + 1,
                            minLength)

            substrings(i) = source.Substring(startIndex, length)
            startIndex += length
        Next

        Return substrings
    End Function

#End Region 'Substring

#Region "Contains"

    ''' <summary>
    ''' Returns a value indicating whether a specified Unicode character occurs within the inpout string.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="value">
    ''' A Unicode character to seek.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if <i>value</i> Is found within the input string; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function Contains(source As String, value As Char) As Boolean
        Return source.IndexOf(value) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether a specified Unicode character occurs within the inpout string. The search starts at a specified character position.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="value">
    ''' A Unicode character to seek.
    ''' </param>
    ''' <param name="startIndex">
    ''' The search starting position.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if <i>value</i> Is found within the input string; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function Contains(source As String,
                             value As Char,
                             startIndex As Integer) As Boolean
        Return source.IndexOf(value, startIndex) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether a specified substring occurs within the inpout string. The search starts at a specified character position.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="value">
    ''' The string to seek.
    ''' </param>
    ''' <param name="startIndex">
    ''' The search starting position.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if <i>value</i> Is found within the input string; otherwise, <b>false</b>. If <i>value</i> Is <see cref="String.Empty"/>, the return value Is <b>true</b>.
    ''' </returns>
    <Extension>
    Public Function Contains(source As String,
                             value As String,
                             startIndex As Integer) As Boolean
        Return source.IndexOf(value, startIndex) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether a specified substring occurs within the inpout string.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="value">
    ''' The string to seek.
    ''' </param>
    ''' <param name="comparisonType">
    ''' One of the enumeration values that specifies the rules for the search.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if <i>value</i> Is found within the input string; otherwise, <b>false</b>. If <i>value</i> Is <see cref="String.Empty"/>, the return value Is <b>true</b>.
    ''' </returns>
    <Extension>
    Public Function Contains(source As String,
                             value As String,
                             comparisonType As StringComparison) As Boolean
        Return source.IndexOf(value, comparisonType) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether a specified Unicode character occurs within the inpout string.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="value">
    ''' A Unicode character to seek.
    ''' </param>
    ''' <param name="startIndex">
    ''' The search starting position.
    ''' </param>
    ''' <param name="count">
    ''' The number of character positions to examine.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if <i>value</i> Is found within the input string; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function Contains(source As String,
                             value As Char,
                             startIndex As Integer,
                             count As Integer) As Boolean
        Return source.IndexOf(value, startIndex, count) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether a specified substring occurs within the inpout string.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="value">
    ''' The string to seek.
    ''' </param>
    ''' <param name="startIndex">
    ''' The search starting position.
    ''' </param>
    ''' <param name="count">
    ''' The number of character positions to examine.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if <i>value</i> Is found within the input string; otherwise, <b>false</b>. If <i>value</i> Is <see cref="String.Empty"/>, the return value Is <b>true</b>.
    ''' </returns>
    <Extension>
    Public Function Contains(source As String,
                             value As String,
                             startIndex As Integer,
                             count As Integer) As Boolean
        Return source.IndexOf(value, startIndex, count) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether a specified substring occurs within the inpout string.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="value">
    ''' The string to seek.
    ''' </param>
    ''' <param name="startIndex">
    ''' The search starting position.
    ''' </param>
    ''' <param name="comparisonType">
    ''' One of the enumeration values that specifies the rules for the search.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if <i>value</i> Is found within the input string; otherwise, <b>false</b>. If <i>value</i> Is <see cref="String.Empty"/>, the return value Is <b>true</b>.
    ''' </returns>
    <Extension>
    Public Function Contains(source As String,
                             value As String,
                             startIndex As Integer,
                             comparisonType As StringComparison) As Boolean
        Return source.IndexOf(value, startIndex, comparisonType) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether a specified substring occurs within the inpout string.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="value">
    ''' The string to seek.
    ''' </param>
    ''' <param name="startIndex">
    ''' The search starting position.
    ''' </param>
    ''' <param name="count">
    ''' The number of character positions to examine.
    ''' </param>
    ''' <param name="comparisonType">
    ''' One of the enumeration values that specifies the rules for the search.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if <i>value</i> Is found within the input string; otherwise, <b>false</b>. If <i>value</i> Is <see cref="String.Empty"/>, the return value Is <b>true</b>.
    ''' </returns>
    <Extension>
    Public Function Contains(source As String,
                             value As String,
                             startIndex As Integer,
                             count As Integer,
                             comparisonType As StringComparison) As Boolean
        Return source.IndexOf(value, startIndex, count, comparisonType) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether any character in a specified array of Unicode characters occur within the inpout string.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="anyOf">
    ''' A Unicode character array containing one Or more characters to seek.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if any character in <i>anyOf</i> Is found within the input string; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function ContainsAny(source As String,
                                anyOf As Char()) As Boolean
        Return source.IndexOfAny(anyOf) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether any character in a specified array of Unicode characters occur within the inpout string.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="anyOf">
    ''' A Unicode character array containing one Or more characters to seek.
    ''' </param>
    ''' <param name="startIndex">
    ''' The search starting position.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if any character in <i>anyOf</i> Is found within the input string; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function ContainsAny(source As String,
                                anyOf As Char(),
                                startIndex As Integer) As Boolean
        Return source.IndexOfAny(anyOf, startIndex) <> -1
    End Function

    ''' <summary>
    ''' Returns a value indicating whether any character in a specified array of Unicode characters occur within the inpout string.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <param name="anyOf">
    ''' A Unicode character array containing one Or more characters to seek.
    ''' </param>
    ''' <param name="startIndex">
    ''' The search starting position.
    ''' </param>
    ''' <param name="count">
    ''' The number of character positions to examine.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if any character in <i>anyOf</i> Is found within the input string; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function ContainsAny(source As String,
                                anyOf As Char(),
                                startIndex As Integer,
                                count As Integer) As Boolean
        Return source.IndexOfAny(anyOf, startIndex, count) <> -1
    End Function

#End Region 'Contains

#Region "Convert to Number by Base"

    ''' <summary>
    ''' Converts a string containing binary digits in a <see cref="Byte"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the binary string.
    ''' </returns>
    <Extension>
    Public Function ToByteFromBinary(source As String) As Byte
        Return Convert.ToByte(source, 2)
    End Function

    ''' <summary>
    ''' Converts a string containing binary digits in a <see cref="Int16"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the binary string.
    ''' </returns>
    <Extension>
    Public Function ToInt16FromBinary(source As String) As Short
        Return Convert.ToInt16(source, 2)
    End Function

    ''' <summary>
    ''' Converts a string containing binary digits in a <see cref="Int32"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the binary string.
    ''' </returns>
    <Extension>
    Public Function ToInt32FromBinary(source As String) As Integer
        Return Convert.ToInt32(source, 2)
    End Function

    ''' <summary>
    ''' Converts a string containing binary digits in a <see cref="Int64"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the binary string.
    ''' </returns>
    <Extension>
    Public Function ToInt64FromBinary(source As String) As Long
        Return Convert.ToInt64(source, 2)
    End Function

    ''' <summary>
    ''' Converts a string containing octal digits in a <see cref="Byte"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the octal string.
    ''' </returns>
    <Extension>
    Public Function ToByteFromOctal(source As String) As Byte
        Return Convert.ToByte(source, 8)
    End Function

    ''' <summary>
    ''' Converts a string containing octal digits in a <see cref="Int16"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the octal string.
    ''' </returns>
    <Extension>
    Public Function ToInt16FromOctal(source As String) As Short
        Return Convert.ToInt16(source, 8)
    End Function

    ''' <summary>
    ''' Converts a string containing octal digits in a <see cref="Int32"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the octal string.
    ''' </returns>
    <Extension>
    Public Function ToInt32FromOctal(source As String) As Integer
        Return Convert.ToInt32(source, 8)
    End Function

    ''' <summary>
    ''' Converts a string containing octal digits in a <see cref="Int64"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the octal string.
    ''' </returns>
    <Extension>
    Public Function ToInt64FromOctal(source As String) As Long
        Return Convert.ToInt64(source, 8)
    End Function

    ''' <summary>
    ''' Converts a string containing decimal digits in a <see cref="Byte"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the decimal string.
    ''' </returns>
    <Extension>
    Public Function ToByteFromDecimal(source As String) As Byte
        Return Convert.ToByte(source, 10)
    End Function

    ''' <summary>
    ''' Converts a string containing decimal digits in a <see cref="Int16"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the decimal string.
    ''' </returns>
    <Extension>
    Public Function ToInt16FromDecimal(source As String) As Short
        Return Convert.ToInt16(source, 10)
    End Function

    ''' <summary>
    ''' Converts a string containing decimal digits in a <see cref="Int32"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the decimal string.
    ''' </returns>
    <Extension>
    Public Function ToInt32FromDecimal(source As String) As Integer
        Return Convert.ToInt32(source, 10)
    End Function

    ''' <summary>
    ''' Converts a string containing decimal digits in a <see cref="Int64"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the decimal string.
    ''' </returns>
    <Extension>
    Public Function ToInt64FromDecimal(source As String) As Long
        Return Convert.ToInt64(source, 10)
    End Function

    ''' <summary>
    ''' Converts a string containing hexadecimal digits in a <see cref="Byte"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the hexadecimal string.
    ''' </returns>
    <Extension>
    Public Function ToByteFromHexadecimal(source As String) As Byte
        Return Convert.ToByte(source, 16)
    End Function

    ''' <summary>
    ''' Converts a string containing hexadecimal digits in a <see cref="Int16"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the hexadecimal string.
    ''' </returns>
    <Extension>
    Public Function ToInt16FromHexadecimal(source As String) As Short
        Return Convert.ToInt16(source, 16)
    End Function

    ''' <summary>
    ''' Converts a string containing hexadecimal digits in a <see cref="Int32"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the hexadecimal string.
    ''' </returns>
    <Extension>
    Public Function ToInt32FromHexadecimal(source As String) As Integer
        Return Convert.ToInt32(source, 16)
    End Function

    ''' <summary>
    ''' Converts a string containing hexadecimal digits in a <see cref="Int64"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The input string.
    ''' </param>
    ''' <returns>
    ''' The number represented by the hexadecimal string.
    ''' </returns>
    <Extension>
    Public Function ToInt64FromHexadecimal(source As String) As Long
        Return Convert.ToInt64(source, 16)
    End Function

#End Region 'Convert to Number by Base

#Region "Numeric Validation"

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="Byte"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="Byte"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsByte(source As String) As Boolean
        Return Byte.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="SByte"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="SByte"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsSByte(source As String) As Boolean
        Return SByte.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="Int16"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="Int16"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsInt16(source As String) As Boolean
        Return Short.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="UInt16"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="UInt16"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsUInt16(source As String) As Boolean
        Return UShort.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="Int32"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="Int32"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsInt32(source As String) As Boolean
        Return Integer.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="UInt32"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="UInt32"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsUInt32(source As String) As Boolean
        Return UInteger.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="Int64"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="Int64"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsInt64(source As String) As Boolean
        Return Long.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="UInt64"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="UInt64"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsUInt64(source As String) As Boolean
        Return ULong.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="Single"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="Single"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsSingle(source As String) As Boolean
        Return Single.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="Double"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="Double"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsDouble(source As String) As Boolean
        Return Double.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="Decimal"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="Decimal"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsDecimal(source As String) As Boolean
        Return Decimal.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="Guid"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="Guid"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsGuid(source As String) As Boolean
        Return Guid.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="Guid"/> from the specified format.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <param name="format">
    ''' One of the following specifiers that indicates the exact format to use when interpreting input: "N", "D", "B", "P", Or "X".
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="Guid"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsGuid(source As String, format As String) As Boolean
        Return Guid.TryParseExact(source, format, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="BigInteger"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="BigInteger"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsBigInteger(source As String) As Boolean
        Return BigInteger.TryParse(source, Nothing)
    End Function

#End Region 'Numeric Validation

#Region "Date/Time Validation"

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="DateTime"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="DateTime"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsDateTime(source As String) As Boolean
        Return Date.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="DateTime"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <param name="format">
    ''' The required format of the input string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="DateTime"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsDateTime(source As String, format As String) As Boolean
        Return Date.TryParseExact(source, format, Nothing, DateTimeStyles.None, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="DateTime"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <param name="formats">
    ''' An array of the allowable formats of the input string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="DateTime"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsDateTime(source As String, ParamArray formats As String()) As Boolean
        Return Date.TryParseExact(source, formats, Nothing, DateTimeStyles.None, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="TimeSpan"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="TimeSpan"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsTimeSpan(source As String) As Boolean
        Return TimeSpan.TryParse(source, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="TimeSpan"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <param name="format">
    ''' The required format of the input string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="TimeSpan"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsTimeSpan(source As String, format As String) As Boolean
        Return TimeSpan.TryParseExact(source, format, Nothing, TimeSpanStyles.None, Nothing)
    End Function

    ''' <summary>
    ''' Determines whether a string can be converted to a <see cref="TimeSpan"/>.
    ''' </summary>
    ''' <param name="source">
    ''' The inout string.
    ''' </param>
    ''' <param name="formats">
    ''' An array of the allowable formats of the input string.
    ''' </param>
    ''' <returns>
    ''' <b>true</b> if the inout string can be converted to a <see cref="TimeSpan"/>; otherwise, <b>false</b>.
    ''' </returns>
    <Extension>
    Public Function IsTimeSpan(source As String, ParamArray formats As String()) As Boolean
        Return TimeSpan.TryParseExact(source, formats, Nothing, TimeSpanStyles.None, Nothing)
    End Function

#End Region 'Date/Time Validation

End Module
