﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace Wunnell.Extensions.Tests.NetFx.CS
{
    [TestClass]
    public class EnumerableExtensionsTest
    {
        #region Randomize

        [TestMethod]
        public void RandomizeThrowsOnNullReference()
        {
            // Arrange
            IEnumerable<object> source = null;
            Exception exception = null;

            // Act
            try
            {
                source.Randomize();
            }
            catch (Exception e)
            {
                exception = e;
            }

            // Assert
            Assert.IsNotNull(exception);
        }

        [TestMethod]
        public void RandomizeReturnsSameItems()
        {
            // Arrange
            var source = Enumerable.Range(1, 10)
                                   .Select(n => new object())
                                   .ToArray();

            // Act
            var result = source.Randomize().ToArray();

            // Assert
            Assert.AreEqual(source.Length, result.Length);
            Assert.IsTrue(source.All(o => result.Contains(o)));
        }

        [TestMethod]
        public void RandomizeChangesOrder()
        {
            // Arrange
            var source = Enumerable.Range(1, 10)
                                   .Select(n => new object())
                                   .ToArray();

            // Act
            var result = source.Randomize().ToArray();

            // Assert
            Assert.IsTrue(Enumerable.Range(0, source.Length).Any(n => source[n] != result[n]));
        }

        #endregion Randomize

        #region Randomize(Random)

        [TestMethod]
        public void RandomizeWithRandomThrowsOnNullReference()
        {
            // Arrange
            IEnumerable<object> source = null;
            var random = new Random();
            Exception exception = null;

            // Act
            try
            {
                source.Randomize(random);
            }
            catch (Exception e)
            {
                exception = e;
            }

            // Assert
            Assert.IsNotNull(exception);
        }

        [TestMethod]
        public void RandomizeWithRandomReturnsSameItems()
        {
            // Arrange
            var source = Enumerable.Range(1, 10)
                                   .Select(n => new object())
                                   .ToArray();
            var random = new Random();

            // Act
            var result = source.Randomize(random).ToArray();

            // Assert
            Assert.AreEqual(source.Length, result.Length);
            Assert.IsTrue(source.All(o => result.Contains(o)));
        }

        [TestMethod]
        public void RandomizeWithRandomChangesOrder()
        {
            // Arrange
            var source = Enumerable.Range(1, 10)
                                   .Select(n => new object())
                                   .ToArray();
            var random = new Random();

            // Act
            var result = source.Randomize(random).ToArray();

            // Assert
            Assert.IsTrue(Enumerable.Range(0, source.Length).Any(n => source[n] != result[n]));
        }

        [TestMethod]
        public void RandomizeWithRandomProducesExpectedOrder()
        {
            // Arrange
            var source = Enumerable.Range(1, 10)
                                   .ToArray();
            var seed = Convert.ToInt32(DateTime.Now.Ticks % int.MaxValue);
            var random1 = new Random(seed);
            var random2 = new Random(seed);

            // Act
            var result = source.Randomize(random1).ToArray();
            var keys = Enumerable.Range(1, source.Length)
                                 .Select(n => random2.NextDouble())
                                 .ToArray();

            Array.Sort(keys, source);

            // Assert
            Assert.IsTrue(Enumerable.Range(0, source.Length).All(n => source[n] == result[n]));
        }

        #endregion Randomize(Random)
    }
}
